import pydantic.version
from packaging.version import Version

_pydantic_version = Version(pydantic.version.VERSION)

PYDANTIC_VERSION_2 = _pydantic_version >= Version("2.0")

# The pydantic.Discriminator type was added in v2.5.0.
# https://docs.pydantic.dev/latest/changelog/#v250-2023-11-13
PYDANTIC_SUPPORTS_CALLABLE_DISCRIMINATOR = _pydantic_version >= Version("2.5.0")

# This can be used to silence deprecation warnings while we migrate.
if PYDANTIC_VERSION_2:
    from pydantic import PydanticDeprecatedSince20  # type: ignore
else:

    class PydanticDeprecatedSince20(Warning):  # type: ignore
        pass


if PYDANTIC_VERSION_2:
    from pydantic import BaseModel as GenericModel
    from pydantic.v1 import (  # type: ignore
        BaseModel as v1_BaseModel,
        Extra as v1_Extra,
        Field as v1_Field,
        root_validator as v1_root_validator,
        validator as v1_validator,
    )
else:
    from pydantic import (  # type: ignore
        BaseModel as v1_BaseModel,
        Extra as v1_Extra,
        Field as v1_Field,
        root_validator as v1_root_validator,
        validator as v1_validator,
    )
    from pydantic.generics import GenericModel  # type: ignore


class v1_ConfigModel(v1_BaseModel):
    """A simplified variant of our main ConfigModel class.

    This one only uses pydantic v1 features.
    """

    class Config:
        extra = v1_Extra.forbid
        underscore_attrs_are_private = True


__all__ = [
    "PYDANTIC_VERSION_2",
    "PYDANTIC_SUPPORTS_CALLABLE_DISCRIMINATOR",
    "PydanticDeprecatedSince20",
    "GenericModel",
    "v1_ConfigModel",
    "v1_Field",
    "v1_root_validator",
    "v1_validator",
]



import logging
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, List, Optional, Union

import cachetools
from pydantic import BaseModel

from datahub.api.entities.platformresource.platform_resource import (
    PlatformResource,
    PlatformResourceKey,
)
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import PlatformResourceUrn, Urn
from datahub.utilities.search_utils import ElasticDocumentQuery

logger = logging.getLogger(__name__)


class PlatformResourceRepository:
    def __init__(self, graph: DataHubGraph):
        self.graph = graph
        self.cache: cachetools.TTLCache = cachetools.TTLCache(maxsize=1000, ttl=60 * 5)

    def search_by_filter(
        self, filter: ElasticDocumentQuery, add_to_cache: bool = True
    ) -> Iterable[PlatformResource]:
        results = PlatformResource.search_by_filters(self.graph, filter)
        for platform_resource in results:
            if add_to_cache:
                self.cache[platform_resource.id] = platform_resource
            yield platform_resource

    def create(self, platform_resource: PlatformResource) -> None:
        platform_resource.to_datahub(self.graph)
        self.cache[platform_resource.id] = platform_resource

    def get(self, key: PlatformResourceKey) -> Optional[PlatformResource]:
        return self.cache.get(key.id)

    def delete(self, key: PlatformResourceKey) -> None:
        self.graph.delete_entity(urn=PlatformResourceUrn(key.id).urn(), hard=True)
        del self.cache[key.id]


class ExternalEntityId:
    """
    ExternalEntityId is a unique
    identifier for an ExternalEntity.
    """

    @abstractmethod
    def to_platform_resource_key(self) -> PlatformResourceKey:
        """
        Converts the ExternalEntityId to a PlatformResourceKey.
        """
        pass


class CaseSensitivity(Enum):
    UPPER = "upper"
    LOWER = "lower"
    MIXED = "mixed"

    @staticmethod
    def detect_case_sensitivity(value: str) -> "CaseSensitivity":
        if value.isupper():
            return CaseSensitivity.UPPER
        elif value.islower():
            return CaseSensitivity.LOWER
        return CaseSensitivity.MIXED

    @staticmethod
    def detect_for_many(values: List[str]) -> "CaseSensitivity":
        """
        Detects the case sensitivity for a list of strings.
        Returns CaseSensitivity.MIXED if the case sensitivity is mixed.
        """
        if len(values) == 0:
            return CaseSensitivity.MIXED

        if all(
            CaseSensitivity.detect_case_sensitivity(value) == CaseSensitivity.UPPER
            for value in values
        ):
            return CaseSensitivity.UPPER
        elif all(
            CaseSensitivity.detect_case_sensitivity(value) == CaseSensitivity.LOWER
            for value in values
        ):
            return CaseSensitivity.LOWER
        return CaseSensitivity.MIXED


class LinkedResourceSet(BaseModel):
    """
    A LinkedResourceSet is a set of DataHub URNs that are linked to an ExternalEntity.
    """

    urns: List[str]

    def _has_conflict(self, urn: Urn) -> bool:
        """
        Detects if the urn is safe to add into the set
        This is used to detect conflicts between DataHub URNs that are linked to
        the same ExternalEntity.
        e.g. Case sensitivity of URNs
        Mixing tags and terms in the same set etc.
        Return True if the urn is not safe to add into the set, else False.
        If the urn is already in the set, we don't need to add it again, but
        that is not a conflict.
        """
        if urn.urn() in self.urns:
            return False

        # Detect the entity_type of the urns in the existing set
        detected_entity_type = None
        for existing_urn in self.urns:
            try:
                parsed_urn = Urn.from_string(existing_urn)
                entity_type = parsed_urn.entity_type
                if detected_entity_type is None:
                    detected_entity_type = entity_type
                elif detected_entity_type != entity_type:
                    logger.warning(
                        f"Detected entity_type {detected_entity_type} is not equals to {entity_type}"
                    )
                    return True
            except ValueError:
                # Not a valid URN
                logger.warning(f"Invalid URN {existing_urn} in LinkedResourceSet")
                return True
        try:
            parsed_urn = urn
            if (
                detected_entity_type is not None
                and parsed_urn.entity_type != detected_entity_type
            ):
                logger.warning(
                    f"Detected entity_type {detected_entity_type} is not equals to parsed_urn's entity_type: {parsed_urn.entity_type}"
                )
                return True
        except ValueError:
            # Not a valid URN
            logger.warning(f"Invalid URN: {urn} in LinkedResourceSet")
            return True
        return False

    def add(self, urn: Union[str, Urn]) -> bool:
        """
        Adds a URN to the set.
        Returns True if the URN was added, False if it was already in the set.
        Raises a ValueError if the URN is in conflict with the existing set.
        """
        # Deduplicate the URNs if we have somehow duplicate items from concurrent runs
        self.urns = list(set(self.urns))
        if isinstance(urn, str):
            urn = Urn.from_string(urn)
        if self._has_conflict(urn):
            raise ValueError(f"Conflict detected when adding URN {urn} to the set")
        if urn.urn() not in self.urns:
            self.urns.append(urn.urn())
            return True
        return False


class ExternalEntity:
    """
    An ExternalEntity is a representation of an entity that external to DataHub
    but could be linked to one or more DataHub entities.
    """

    @abstractmethod
    def is_managed_by_datahub(self) -> bool:
        """
        Returns whether the entity is managed by DataHub.
        """
        pass

    @abstractmethod
    def datahub_linked_resources(self) -> LinkedResourceSet:
        """
        Returns the URNs of the DataHub entities linked to the external entity.
        Empty list if no linked entities.
        """
        pass

    @abstractmethod
    def as_platform_resource(self) -> PlatformResource:
        """
        Converts the ExternalEntity to a PlatformResource.
        """
        pass

    @abstractmethod
    def get_id(self) -> ExternalEntityId:
        """
        Returns the ExternalEntityId for the ExternalEntity.
        """
        pass


@dataclass
class MissingExternalEntity(ExternalEntity):
    id: ExternalEntityId

    def is_managed_by_datahub(self) -> bool:
        return False

    def datahub_linked_resources(self) -> LinkedResourceSet:
        return LinkedResourceSet(urns=[])

    def as_platform_resource(self) -> Optional[PlatformResource]:  # type: ignore[override]
        return None

    def get_id(self) -> ExternalEntityId:
        return self.id


class ExternalSystem:
    @abstractmethod
    def exists(self, external_entity_id: ExternalEntityId) -> bool:
        """
        Returns whether the ExternalEntityId exists in the external system.
        """
        pass

    @abstractmethod
    def get(
        self,
        external_entity_id: ExternalEntityId,
        platform_resource_repository: PlatformResourceRepository,
    ) -> Optional[ExternalEntity]:
        """
        Returns the ExternalEntity for the ExternalEntityId.
        Uses the platform resource repository to enrich the ExternalEntity with DataHub URNs.
        """
        pass



import dataclasses
import logging
import warnings
from typing import Dict, Optional

import pydantic
from pydantic import Field, root_validator
from typing_extensions import Literal

from datahub.configuration.common import (
    AllowDenyPattern,
    ConfigModel,
    ConfigurationWarning,
)
from datahub.configuration.source_common import DatasetSourceConfigMixin
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.emitter.mce_builder import DEFAULT_ENV
from datahub.ingestion.api.report import Report
from datahub.ingestion.source.bigquery_v2.bigquery_connection import (
    BigQueryConnectionConfig,
)
from datahub.ingestion.source.snowflake.snowflake_connection import (
    SnowflakeConnectionConfig,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.utilities.lossy_collections import LossyList
from datahub.utilities.perf_timer import PerfTimer

logger = logging.getLogger(__name__)


class Constant:
    """
    keys used in fivetran plugin
    """

    ORCHESTRATOR = "fivetran"
    # table column name
    SOURCE_SCHEMA_NAME = "source_schema_name"
    SOURCE_TABLE_NAME = "source_table_name"
    SOURCE_TABLE_ID = "source_table_id"
    SOURCE_COLUMN_NAME = "source_column_name"
    DESTINATION_SCHEMA_NAME = "destination_schema_name"
    DESTINATION_TABLE_NAME = "destination_table_name"
    DESTINATION_TABLE_ID = "destination_table_id"
    DESTINATION_COLUMN_NAME = "destination_column_name"
    SYNC_ID = "sync_id"
    MESSAGE_DATA = "message_data"
    TIME_STAMP = "time_stamp"
    STATUS = "status"
    USER_ID = "user_id"
    EMAIL = "email"
    CONNECTOR_ID = "connector_id"
    CONNECTOR_NAME = "connector_name"
    CONNECTOR_TYPE_ID = "connector_type_id"
    PAUSED = "paused"
    SYNC_FREQUENCY = "sync_frequency"
    DESTINATION_ID = "destination_id"
    CONNECTING_USER_ID = "connecting_user_id"
    # Job status constants
    SUCCESSFUL = "SUCCESSFUL"
    FAILURE_WITH_TASK = "FAILURE_WITH_TASK"
    CANCELED = "CANCELED"


KNOWN_DATA_PLATFORM_MAPPING = {
    "postgres": "postgres",
    "snowflake": "snowflake",
}


class SnowflakeDestinationConfig(SnowflakeConnectionConfig):
    database: str = Field(description="The fivetran connector log database.")
    log_schema: str = Field(description="The fivetran connector log schema.")


class BigQueryDestinationConfig(BigQueryConnectionConfig):
    dataset: str = Field(description="The fivetran connector log dataset.")


class FivetranLogConfig(ConfigModel):
    destination_platform: Literal["snowflake", "bigquery"] = pydantic.Field(
        default="snowflake",
        description="The destination platform where fivetran connector log tables are dumped.",
    )
    snowflake_destination_config: Optional[SnowflakeDestinationConfig] = pydantic.Field(
        default=None,
        description="If destination platform is 'snowflake', provide snowflake configuration.",
    )
    bigquery_destination_config: Optional[BigQueryDestinationConfig] = pydantic.Field(
        default=None,
        description="If destination platform is 'bigquery', provide bigquery configuration.",
    )
    _rename_destination_config = pydantic_renamed_field(
        "destination_config", "snowflake_destination_config"
    )

    @root_validator(pre=True)
    def validate_destination_platfrom_and_config(cls, values: Dict) -> Dict:
        destination_platform = values["destination_platform"]
        if destination_platform == "snowflake":
            if "snowflake_destination_config" not in values:
                raise ValueError(
                    "If destination platform is 'snowflake', user must provide snowflake destination configuration in the recipe."
                )
        elif destination_platform == "bigquery":
            if "bigquery_destination_config" not in values:
                raise ValueError(
                    "If destination platform is 'bigquery', user must provide bigquery destination configuration in the recipe."
                )
        else:
            raise ValueError(
                f"Destination platform '{destination_platform}' is not yet supported."
            )
        return values


@dataclasses.dataclass
class MetadataExtractionPerfReport(Report):
    connectors_metadata_extraction_sec: PerfTimer = dataclasses.field(
        default_factory=PerfTimer
    )
    connectors_lineage_extraction_sec: PerfTimer = dataclasses.field(
        default_factory=PerfTimer
    )
    connectors_jobs_extraction_sec: PerfTimer = dataclasses.field(
        default_factory=PerfTimer
    )


@dataclasses.dataclass
class FivetranSourceReport(StaleEntityRemovalSourceReport):
    connectors_scanned: int = 0
    filtered_connectors: LossyList[str] = dataclasses.field(default_factory=LossyList)
    metadata_extraction_perf: MetadataExtractionPerfReport = dataclasses.field(
        default_factory=MetadataExtractionPerfReport
    )

    def report_connectors_scanned(self, count: int = 1) -> None:
        self.connectors_scanned += count

    def report_connectors_dropped(self, connector: str) -> None:
        self.filtered_connectors.append(connector)


class PlatformDetail(ConfigModel):
    platform: Optional[str] = pydantic.Field(
        default=None,
        description="Override the platform type detection.",
    )
    platform_instance: Optional[str] = pydantic.Field(
        default=None,
        description="The instance of the platform that all assets produced by this recipe belong to",
    )
    env: str = pydantic.Field(
        default=DEFAULT_ENV,
        description="The environment that all assets produced by DataHub platform ingestion source belong to",
    )
    database: Optional[str] = pydantic.Field(
        default=None,
        description="The database that all assets produced by this connector belong to. "
        "For destinations, this defaults to the fivetran log config's database.",
    )
    include_schema_in_urn: bool = pydantic.Field(
        default=True,
        description="Include schema in the dataset URN. In some cases, the schema is not relevant to the dataset URN and Fivetran sets it to the source and destination table names in the connector.",
    )


class FivetranSourceConfig(StatefulIngestionConfigBase, DatasetSourceConfigMixin):
    fivetran_log_config: FivetranLogConfig = pydantic.Field(
        description="Fivetran log connector destination server configurations.",
    )
    connector_patterns: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Filtering regex patterns for connector names.",
    )
    destination_patterns: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for destination ids to filter in ingestion. "
        "Fivetran destination IDs are usually two word identifiers e.g. canyon_tolerable, and are not the same as the destination database name. "
        "They're visible in the Fivetran UI under Destinations -> Overview -> Destination Group ID.",
    )
    include_column_lineage: bool = Field(
        default=True,
        description="Populates table->table column lineage.",
    )

    # Configuration for stateful ingestion
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = pydantic.Field(
        default=None, description="Airbyte Stateful Ingestion Config."
    )

    # Fivetran connector all sources to platform instance mapping
    sources_to_platform_instance: Dict[str, PlatformDetail] = pydantic.Field(
        default={},
        description="A mapping from connector id to its platform/instance/env/database details.",
    )
    # Fivetran destination to platform instance mapping
    destination_to_platform_instance: Dict[str, PlatformDetail] = pydantic.Field(
        default={},
        description="A mapping of destination id to its platform/instance/env details.",
    )

    @pydantic.root_validator(pre=True)
    def compat_sources_to_database(cls, values: Dict) -> Dict:
        if "sources_to_database" in values:
            warnings.warn(
                "The sources_to_database field is deprecated, please use sources_to_platform_instance instead.",
                ConfigurationWarning,
                stacklevel=2,
            )
            mapping = values.pop("sources_to_database")

            values.setdefault("sources_to_platform_instance", {})
            for key, value in mapping.items():
                values["sources_to_platform_instance"].setdefault(key, {})
                values["sources_to_platform_instance"][key].setdefault(
                    "database", value
                )

        return values

    history_sync_lookback_period: int = pydantic.Field(
        7,
        description="The number of days to look back when extracting connectors' sync history.",
    )



import logging
import os
from pathlib import Path
from typing import Dict, List, Optional

import click
from click_default_group import DefaultGroup

from datahub.api.entities.assertion.assertion_config_spec import AssertionsConfigSpec
from datahub.api.entities.assertion.compiler_interface import (
    AssertionCompilationResult,
    CompileResultArtifact,
    CompileResultArtifactType,
)
from datahub.emitter.mce_builder import make_assertion_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import get_default_graph
from datahub.ingestion.graph.config import ClientMode
from datahub.integrations.assertion.registry import ASSERTION_PLATFORMS
from datahub.upgrade import upgrade

logger = logging.getLogger(__name__)

REPORT_FILE_NAME = "compile_report.json"


@click.group(cls=DefaultGroup, default="upsert")
def assertions() -> None:
    """A group of commands to interact with the Assertion entity in DataHub."""
    pass


@assertions.command()
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@upgrade.check_upgrade
def upsert(file: str) -> None:
    """Upsert (create or update) a set of assertions in DataHub."""

    assertions_spec: AssertionsConfigSpec = AssertionsConfigSpec.from_yaml(file)

    with get_default_graph(ClientMode.CLI) as graph:
        for assertion_spec in assertions_spec.assertions:
            try:
                mcp = MetadataChangeProposalWrapper(
                    entityUrn=make_assertion_urn(assertion_spec.get_id()),
                    aspect=assertion_spec.get_assertion_info_aspect(),
                )
                graph.emit_mcp(mcp)
                # TODO: Validate uniqueness of assertion ids. Report if duplicates found.
                # TODO: Use upsert graphql endpoints here instead of graph.emit_mcp.
                click.secho(f"Update succeeded for urn {mcp.entityUrn}.", fg="green")
            except Exception as e:
                logger.exception(e)
                click.secho(
                    f"Update failed for {mcp.entityUrn}: {e}",
                    fg="red",
                )


@assertions.command()
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@click.option("-p", "--platform", required=True, type=str)
@click.option("-o", "--output-to", required=False, type=click.Path(exists=True))
@click.option(
    "-x",
    "--extras",
    required=False,
    multiple=True,
    default=[],
    help="Platform-specific extra key-value inputs in form key=value",
)
def compile(
    file: str, platform: str, output_to: Optional[str], extras: List[str]
) -> None:
    """Compile a set of assertions for input assertion platform.
    Note that this does not run any code or execute any queries on assertion platform
    and only creates artifacts specific to assertion platform that can be executed manually.
    In future, we may introduce separate command to automatically apply these compiled changes
    in assertion platform. Currently, generated result artifacts are stored in target folder
    unless another folder is specified using option `--output-to <folder>`.
    """

    if platform not in ASSERTION_PLATFORMS:
        click.secho(
            f"Platform {platform} is not supported.",
            fg="red",
        )

    if output_to is None:
        output_to = f"{os.getcwd()}/target"

    if not os.path.isdir(output_to):
        os.mkdir(output_to)

    assertions_spec: AssertionsConfigSpec = AssertionsConfigSpec.from_yaml(file)

    try:
        compiler = ASSERTION_PLATFORMS[platform].create(
            output_dir=output_to, extras=extras_list_to_dict(extras)
        )
        result = compiler.compile(assertions_spec)

        write_report_file(output_to, result)
        click.secho("Compile report:", bold=True)
        click.echo(result.report.as_string())
        if result.status == "failure":
            click.secho("Failure", fg="yellow", bold=True)
        else:
            click.secho("Success", fg="green", bold=True)
    except Exception as e:
        logger.exception(e)
        click.secho(
            f"Compile failed: {e}",
            fg="red",
        )


def write_report_file(output_to: str, result: AssertionCompilationResult) -> None:
    report_path = Path(output_to) / REPORT_FILE_NAME
    with (report_path).open("w") as f:
        result.add_artifact(
            CompileResultArtifact(
                name=REPORT_FILE_NAME,
                path=report_path,
                type=CompileResultArtifactType.COMPILE_REPORT,
                description="Detailed report about compile status",
            )
        )
        f.write(result.report.as_json())


def extras_list_to_dict(extras: List[str]) -> Dict[str, str]:
    extra_properties: Dict[str, str] = dict()
    for x in extras:
        parts = x.split("=")
        assert len(parts) == 2, (
            f"Invalid value for extras {x}, should be in format key=value"
        )
        extra_properties[parts[0]] = parts[1]
    return extra_properties


# TODO: support for
# Immediate:
# 1. delete assertions (from datahub)
# Later:
# 3. execute compiled assertions on assertion platform (Later, requires connection details to platform),
# 4. cleanup assertions from assertion platform (generate artifacts. optionally execute)



from typing import Any, Dict, Optional

from sqlalchemy.engine import URL


def make_sqlalchemy_uri(
    scheme: str,
    username: Optional[str],
    password: Optional[str],
    at: Optional[str],
    db: Optional[str],
    uri_opts: Optional[Dict[str, Any]] = None,
) -> str:
    host: Optional[str] = None
    port: Optional[int] = None
    if at:
        try:
            host, port_str = at.rsplit(":", 1)
            port = int(port_str)
        except ValueError:
            host = at
            port = None
    if uri_opts:
        uri_opts = {k: v for k, v in uri_opts.items() if v is not None}

    return str(
        URL.create(
            drivername=scheme,
            username=username,
            password=password,
            host=host,
            port=port,
            database=db,
            query=uri_opts or {},
        )
    )



from typing import List, Optional, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    FormInfoClass as FormInfo,
    FormPromptClass,
    KafkaAuditHeaderClass,
    SystemMetadataClass,
)
from datahub.specific.aspect_helpers.ownership import HasOwnershipPatch
from datahub.utilities.urns.urn import Urn


class FormPatchBuilder(HasOwnershipPatch, MetadataPatchProposal):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        super().__init__(
            urn, system_metadata=system_metadata, audit_header=audit_header
        )

    def set_name(self, name: Optional[str] = None) -> "FormPatchBuilder":
        if name is not None:
            self._add_patch(
                FormInfo.ASPECT_NAME,
                "add",
                path=("name",),
                value=name,
            )
        return self

    def set_description(self, description: Optional[str] = None) -> "FormPatchBuilder":
        if description is not None:
            self._add_patch(
                FormInfo.ASPECT_NAME,
                "add",
                path=("description",),
                value=description,
            )
        return self

    def set_type(self, type: Optional[str] = None) -> "FormPatchBuilder":
        if type is not None:
            self._add_patch(
                FormInfo.ASPECT_NAME,
                "add",
                path=("type",),
                value=type,
            )
        return self

    def add_prompt(self, prompt: FormPromptClass) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "add",
            path=("prompts", prompt.id),
            value=prompt,
        )
        return self

    def add_prompts(self, prompts: List[FormPromptClass]) -> "FormPatchBuilder":
        for prompt in prompts:
            self.add_prompt(prompt)
        return self

    def remove_prompt(self, prompt_id: str) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "remove",
            path=("prompts", prompt_id),
            value=prompt_id,
        )
        return self

    def remove_prompts(self, prompt_ids: List[str]) -> "FormPatchBuilder":
        for prompt_id in prompt_ids:
            self.remove_prompt(prompt_id)
        return self

    def set_ownership_form(self, is_ownership: bool) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "add",
            path=("actors", "owners"),
            value=is_ownership,
        )
        return self

    def add_assigned_user(self, user_urn: Union[str, Urn]) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "add",
            path=("actors", "users", user_urn),
            value=user_urn,
        )
        return self

    def remove_assigned_user(self, user_urn: Union[str, Urn]) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "remove",
            path=("actors", "users", user_urn),
            value=user_urn,
        )
        return self

    def add_assigned_group(self, group_urn: Union[str, Urn]) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "add",
            path=("actors", "groups", group_urn),
            value=group_urn,
        )
        return self

    def remove_assigned_group(self, group_urn: Union[str, Urn]) -> "FormPatchBuilder":
        self._add_patch(
            FormInfo.ASPECT_NAME,
            "remove",
            path=("actors", "groups", group_urn),
            value=group_urn,
        )
        return self



import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional

import numpy as np
from cassandra.util import OrderedMapSerializedKey, SortedSet

from datahub.emitter.mce_builder import make_dataset_urn_with_platform_instance
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.cassandra.cassandra_api import (
    CassandraAPI,
    CassandraColumn,
    CassandraEntities,
    CassandraQueries,
)
from datahub.ingestion.source.cassandra.cassandra_config import CassandraSourceConfig
from datahub.ingestion.source.cassandra.cassandra_utils import CassandraSourceReport
from datahub.ingestion.source_report.ingestion_stage import PROFILING
from datahub.metadata.schema_classes import (
    DatasetFieldProfileClass,
    DatasetProfileClass,
    QuantileClass,
)

logger = logging.getLogger(__name__)


@dataclass
class ColumnMetric:
    col_type: str = ""
    values: List[Any] = field(default_factory=list)
    null_count: int = 0
    total_count: int = 0
    distinct_count: Optional[int] = None
    min: Optional[Any] = None
    max: Optional[Any] = None
    mean: Optional[float] = None
    stdev: Optional[float] = None
    median: Optional[float] = None
    quantiles: Optional[List[float]] = None
    sample_values: Optional[Any] = None


@dataclass
class ProfileData:
    row_count: Optional[int] = None
    column_count: Optional[int] = None
    column_metrics: Dict[str, ColumnMetric] = field(default_factory=dict)


class CassandraProfiler:
    config: CassandraSourceConfig
    report: CassandraSourceReport

    def __init__(
        self,
        config: CassandraSourceConfig,
        report: CassandraSourceReport,
        api: CassandraAPI,
    ) -> None:
        self.api = api
        self.config = config
        self.report = report

    def get_workunits(
        self, cassandra_data: CassandraEntities
    ) -> Iterable[MetadataWorkUnit]:
        for keyspace_name in cassandra_data.keyspaces:
            tables = cassandra_data.tables.get(keyspace_name, [])
            with (
                self.report.new_stage(f"{keyspace_name}: {PROFILING}"),
                ThreadPoolExecutor(
                    max_workers=self.config.profiling.max_workers
                ) as executor,
            ):
                future_to_dataset = {
                    executor.submit(
                        self.generate_profile,
                        keyspace_name,
                        table_name,
                        cassandra_data.columns.get(table_name, []),
                    ): table_name
                    for table_name in tables
                }
                for future in as_completed(future_to_dataset):
                    table_name = future_to_dataset[future]
                    try:
                        yield from future.result()
                    except Exception as exc:
                        self.report.profiling_skipped_other[table_name] += 1
                        self.report.failure(
                            message="Failed to profile for table",
                            context=f"{keyspace_name}.{table_name}",
                            exc=exc,
                        )

    def generate_profile(
        self,
        keyspace_name: str,
        table_name: str,
        columns: List[CassandraColumn],
    ) -> Iterable[MetadataWorkUnit]:
        dataset_name: str = f"{keyspace_name}.{table_name}"
        dataset_urn = make_dataset_urn_with_platform_instance(
            platform="cassandra",
            name=dataset_name,
            env=self.config.env,
            platform_instance=self.config.platform_instance,
        )

        if not columns:
            self.report.warning(
                message="Skipping profiling as no columns found for table",
                context=f"{keyspace_name}.{table_name}",
            )
            self.report.profiling_skipped_other[table_name] += 1
            return

        if not self.config.profile_pattern.allowed(f"{keyspace_name}.{table_name}"):
            self.report.profiling_skipped_table_profile_pattern[keyspace_name] += 1
            logger.info(
                f"Table {table_name} in {keyspace_name}, not allowed for profiling"
            )
            return

        try:
            profile_data = self.profile_table(keyspace_name, table_name, columns)
        except Exception as e:
            self.report.warning(
                message="Profiling Failed",
                context=f"{keyspace_name}.{table_name}",
                exc=e,
            )
            return

        profile_aspect = self.populate_profile_aspect(profile_data)

        if profile_aspect:
            self.report.report_entity_profiled(table_name)
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn, aspect=profile_aspect
            )
            yield mcp.as_workunit()

    def populate_profile_aspect(self, profile_data: ProfileData) -> DatasetProfileClass:
        field_profiles = [
            self._create_field_profile(column_name, column_metrics)
            for column_name, column_metrics in profile_data.column_metrics.items()
        ]
        return DatasetProfileClass(
            timestampMillis=round(time.time() * 1000),
            rowCount=profile_data.row_count,
            columnCount=profile_data.column_count,
            fieldProfiles=field_profiles,
        )

    def _create_field_profile(
        self, field_name: str, field_stats: ColumnMetric
    ) -> DatasetFieldProfileClass:
        quantiles = field_stats.quantiles
        return DatasetFieldProfileClass(
            fieldPath=field_name,
            uniqueCount=field_stats.distinct_count,
            nullCount=field_stats.null_count,
            min=str(field_stats.min) if field_stats.min else None,
            max=str(field_stats.max) if field_stats.max else None,
            mean=str(field_stats.mean) if field_stats.mean else None,
            median=str(field_stats.median) if field_stats.median else None,
            stdev=str(field_stats.stdev) if field_stats.stdev else None,
            quantiles=[
                QuantileClass(quantile=str(0.25), value=str(quantiles[0])),
                QuantileClass(quantile=str(0.75), value=str(quantiles[1])),
            ]
            if quantiles
            else None,
            sampleValues=field_stats.sample_values
            if field_stats.sample_values
            else None,
        )

    def profile_table(
        self, keyspace_name: str, table_name: str, columns: List[CassandraColumn]
    ) -> ProfileData:
        profile_data = ProfileData()

        resp = self.api.execute(
            CassandraQueries.ROW_COUNT.format(keyspace_name, table_name)
        )
        if resp:
            profile_data.row_count = resp[0].row_count

        profile_data.column_count = len(columns)

        if not self.config.profiling.profile_table_level_only:
            resp = self.api.execute(
                f'SELECT {", ".join([col.column_name for col in columns])} FROM {keyspace_name}."{table_name}"'
            )
            profile_data.column_metrics = self._collect_column_data(resp, columns)

        return self._parse_profile_results(profile_data)

    def _parse_profile_results(self, profile_data: ProfileData) -> ProfileData:
        for cl_name, column_metrics in profile_data.column_metrics.items():
            if column_metrics.values:
                try:
                    self._compute_field_statistics(column_metrics)
                except Exception as e:
                    self.report.warning(
                        message="Profiling Failed For Column Stats",
                        context=cl_name,
                        exc=e,
                    )
                    raise e
        return profile_data

    def _collect_column_data(
        self, rows: List[Any], columns: List[CassandraColumn]
    ) -> Dict[str, ColumnMetric]:
        metrics = {column.column_name: ColumnMetric() for column in columns}

        for row in rows:
            for column in columns:
                if self._is_skippable_type(column.type):
                    continue

                value: Any = getattr(row, column.column_name, None)
                metric = metrics[column.column_name]
                metric.col_type = column.type

                metric.total_count += 1
                if value is None:
                    metric.null_count += 1
                else:
                    metric.values.extend(self._parse_value(value))

        return metrics

    def _is_skippable_type(self, data_type: str) -> bool:
        return data_type.lower() in ["timeuuid", "blob", "frozen<tuple<tinyint, text>>"]

    def _parse_value(self, value: Any) -> List[Any]:
        if isinstance(value, SortedSet):
            return list(value)
        elif isinstance(value, OrderedMapSerializedKey):
            return list(dict(value).values())
        elif isinstance(value, list):
            return value
        return [value]

    def _compute_field_statistics(self, column_metrics: ColumnMetric) -> None:
        values = column_metrics.values
        if not values:
            return

        # ByDefault Null count is added
        if not self.config.profiling.include_field_null_count:
            column_metrics.null_count = 0

        if self.config.profiling.include_field_distinct_count:
            column_metrics.distinct_count = len(set(values))

        if self.config.profiling.include_field_min_value:
            column_metrics.min = min(values)

        if self.config.profiling.include_field_max_value:
            column_metrics.max = max(values)

        if values and self._is_numeric_type(column_metrics.col_type):
            if self.config.profiling.include_field_mean_value:
                column_metrics.mean = round(float(np.mean(values)), 2)
            if self.config.profiling.include_field_stddev_value:
                column_metrics.stdev = round(float(np.std(values)), 2)
            if self.config.profiling.include_field_median_value:
                column_metrics.median = round(float(np.median(values)), 2)
            if self.config.profiling.include_field_quantiles:
                column_metrics.quantiles = [
                    float(np.percentile(values, 25)),
                    float(np.percentile(values, 75)),
                ]

        if values and self.config.profiling.include_field_sample_values:
            column_metrics.sample_values = [str(v) for v in values[:5]]

    def _is_numeric_type(self, data_type: str) -> bool:
        return data_type.lower() in [
            "int",
            "counter",
            "bigint",
            "float",
            "double",
            "decimal",
            "smallint",
            "tinyint",
            "varint",
        ]



import collections
import dataclasses
import json
import logging
from datetime import datetime
from email.utils import parseaddr
from typing import Any, Dict, Iterable, List, Optional

from dateutil import parser
from pydantic.fields import Field
from pydantic.main import BaseModel
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine

import datahub.emitter.mce_builder as builder
from datahub.configuration.time_window_config import get_time_bucket
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.trino import TrinoConfig
from datahub.ingestion.source.usage.usage_common import (
    BaseUsageConfig,
    GenericAggregatedDataset,
)

logger = logging.getLogger(__name__)

trino_datetime_format = "%Y-%m-%d %H:%M:%S.%f %Z"

# Qeurying Starburst completed queries table
# https://docs.starburst.io/latest/security/event-logger.html#completed-queries
trino_usage_sql_comment = """
SELECT DISTINCT usr,
        query,
        "catalog",
        "schema",
        query_type,
        accessed_metadata,
        create_time,
        end_time
FROM {audit_catalog}.{audit_schema}.completed_queries
WHERE 1 = 1
AND query_type  = 'SELECT'
AND create_time >= timestamp '{start_time}'
AND end_time < timestamp '{end_time}'
AND query_state  = 'FINISHED'
ORDER BY end_time desc
""".strip()

TrinoTableRef = str
AggregatedDataset = GenericAggregatedDataset[TrinoTableRef]


class TrinoConnectorInfo(BaseModel):
    partitionIds: List[str]
    truncated: Optional[bool]


class TrinoAccessedMetadata(BaseModel):
    catalog_name: Optional[str] = Field(None, alias="catalogName")
    schema_name: Optional[str] = Field(None, alias="schema")
    table: Optional[str] = None
    columns: List[str]
    connector_info: Optional[TrinoConnectorInfo] = Field(None, alias="connectorInfo")


class TrinoJoinedAccessEvent(BaseModel):
    usr: Optional[str] = None
    query: Optional[str] = None
    catalog: Optional[str] = None
    schema_name: Optional[str] = Field(None, alias="schema")
    query_type: Optional[str] = None
    table: Optional[str] = None
    accessed_metadata: List[TrinoAccessedMetadata]
    starttime: datetime = Field(alias="create_time")
    endtime: Optional[datetime] = Field(alias="end_time")


class EnvBasedSourceBaseConfig:
    pass


class TrinoUsageConfig(TrinoConfig, BaseUsageConfig, EnvBasedSourceBaseConfig):
    email_domain: str = Field(
        description="The email domain which will be appended to the users "
    )
    audit_catalog: str = Field(
        description="The catalog name where the audit table can be found "
    )
    audit_schema: str = Field(
        description="The schema name where the audit table can be found"
    )
    options: dict = Field(default={}, description="")
    database: str = Field(description="The name of the catalog from getting the usage")

    def get_sql_alchemy_url(
        self, uri_opts: Optional[Dict[str, Any]] = None, database: Optional[str] = None
    ) -> str:
        return super().get_sql_alchemy_url(uri_opts=uri_opts, database=database)


@dataclasses.dataclass
class TrinoUsageReport(SourceReport):
    num_joined_access_events_skipped: int = 0


@platform_name("Trino")
@config_class(TrinoUsageConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.USAGE_STATS, "Enabled by default to get usage stats")
@dataclasses.dataclass
class TrinoUsageSource(Source):
    """
    If you are using Starburst Trino you can collect usage stats the following way.

    #### Prerequsities
    1. You need to setup Event Logger which saves audit logs into a Postgres db and setup this db as a catalog in Trino

    2. Install starbust-trino-usage plugin
    Run pip install 'acryl-datahub[starburst-trino-usage]'.

    """

    config: TrinoUsageConfig
    report: TrinoUsageReport = dataclasses.field(default_factory=TrinoUsageReport)

    @classmethod
    def create(cls, config_dict, ctx):
        config = TrinoUsageConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        access_events = self._get_trino_history()
        # If the query results is empty, we don't want to proceed
        if not access_events:
            return

        joined_access_event = self._get_joined_access_event(access_events)
        aggregated_info = self._aggregate_access_events(joined_access_event)

        for time_bucket in aggregated_info.values():
            for aggregate in time_bucket.values():
                yield self._make_usage_stat(aggregate)

    def _make_usage_query(self) -> str:
        return trino_usage_sql_comment.format(
            audit_catalog=self.config.audit_catalog,
            audit_schema=self.config.audit_schema,
            start_time=self.config.start_time.strftime(trino_datetime_format),
            end_time=self.config.end_time.strftime(trino_datetime_format),
        )

    def _make_sql_engine(self) -> Engine:
        url = self.config.get_sql_alchemy_url()
        logger.debug(f"sql_alchemy_url = {url}")
        engine = create_engine(url, **self.config.options)
        return engine

    def _get_trino_history(self):
        query = self._make_usage_query()
        engine = self._make_sql_engine()
        results = engine.execute(query)
        events = []
        for row in results:
            event_dict = row._asdict()

            # stripping extra spaces caused by above _asdict() conversion
            for k, v in event_dict.items():
                if isinstance(v, str):
                    event_dict[k] = v.strip()

            if event_dict.get("starttime", None):
                event_dict["starttime"] = event_dict.get("starttime").__str__()
            if event_dict.get("endtime", None):
                event_dict["endtime"] = event_dict.get("endtime").__str__()

            logger.debug(f"event_dict: {event_dict}")
            events.append(event_dict)

        if events:
            return events

        # SQL results can be empty. If results is empty, the SQL connection closes.
        # Then, we don't want to proceed ingestion.
        logging.info("SQL Result is empty")
        return None

    def _convert_str_to_datetime(self, v):
        if isinstance(v, str):
            isodate = parser.parse(v)  # compatible with Python 3.6+
            return isodate
        if isinstance(v, datetime):
            return v

    def _get_joined_access_event(self, events):
        joined_access_events = []
        for event_dict in events:
            if event_dict.get("create_time"):
                event_dict["create_time"] = self._convert_str_to_datetime(
                    event_dict["create_time"]
                )
            else:
                self.report.num_joined_access_events_skipped += 1
                logging.info("The create_time parameter is missing. Skipping ....")
                continue

            event_dict["end_time"] = self._convert_str_to_datetime(
                event_dict.get("end_time")
            )

            if not event_dict["accessed_metadata"]:
                self.report.num_joined_access_events_skipped += 1
                logging.info("Field accessed_metadata is empty. Skipping ....")
                continue

            event_dict["accessed_metadata"] = json.loads(
                event_dict["accessed_metadata"]
            )

            if not event_dict.get("usr"):
                self.report.num_joined_access_events_skipped += 1
                logging.info("The username parameter is missing. Skipping ....")
                continue

            try:
                joined_access_events.append(TrinoJoinedAccessEvent(**event_dict))
            except Exception as e:
                self.report.num_joined_access_events_skipped += 1
                logger.info(f"Error while parsing TrinoJoinedAccessEvent: {e}")

        return joined_access_events

    def _aggregate_access_events(
        self, events: List[TrinoJoinedAccessEvent]
    ) -> Dict[datetime, Dict[TrinoTableRef, AggregatedDataset]]:
        datasets: Dict[datetime, Dict[TrinoTableRef, AggregatedDataset]] = (
            collections.defaultdict(dict)
        )

        for event in events:
            floored_ts = get_time_bucket(event.starttime, self.config.bucket_duration)
            for metadata in event.accessed_metadata:
                # Skipping queries starting with $system@
                if metadata.catalog_name and metadata.catalog_name.startswith(
                    "$system@"
                ):
                    logging.debug(
                        f"Skipping system query for {metadata.catalog_name}..."
                    )
                    continue

                # Filtering down queries to the selected catalog
                if metadata.catalog_name != self.config.database:
                    continue

                resource = (
                    f"{metadata.catalog_name}.{metadata.schema_name}.{metadata.table}"
                )

                agg_bucket = datasets[floored_ts].setdefault(
                    resource,
                    AggregatedDataset(
                        bucket_start_time=floored_ts,
                        resource=resource,
                    ),
                )

                # add @unknown.com to username
                # current limitation in user stats UI, we need to provide email to show users
                if event.usr and "@" in parseaddr(event.usr)[1]:
                    username = event.usr
                else:
                    username = f"{event.usr if event.usr else 'unknown'}@{self.config.email_domain}"

                agg_bucket.add_read_entry(
                    username,
                    event.query,
                    metadata.columns,
                    user_email_pattern=self.config.user_email_pattern,
                )
        return datasets

    def _make_usage_stat(self, agg: AggregatedDataset) -> MetadataWorkUnit:
        return agg.make_usage_workunit(
            self.config.bucket_duration,
            lambda resource: builder.make_dataset_urn_with_platform_instance(
                "trino",
                resource.lower(),
                self.config.platform_instance,
                self.config.env,
            ),
            self.config.top_n_queries,
            self.config.format_sql_queries,
            self.config.include_top_n_queries,
            self.config.queries_character_limit,
        )

    def get_report(self) -> SourceReport:
        return self.report



import difflib
import json
import logging
import os
import pathlib
import sys
from pathlib import Path
from shutil import copyfile
from tempfile import NamedTemporaryFile
from typing import Optional

import click
from click_default_group import DefaultGroup

from datahub.api.entities.dataproduct.dataproduct import DataProduct
from datahub.cli.specific.file_loader import load_file
from datahub.emitter.mce_builder import (
    make_group_urn,
    make_user_urn,
    validate_ownership_type,
)
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.ingestion.graph.config import ClientMode
from datahub.metadata.schema_classes import OwnerClass, OwnershipTypeClass
from datahub.specific.dataproduct import DataProductPatchBuilder
from datahub.upgrade import upgrade
from datahub.utilities.urns.urn import Urn

logger = logging.getLogger(__name__)


def _get_owner_urn(maybe_urn: str) -> str:
    if make_user_urn(maybe_urn) == maybe_urn or make_group_urn(maybe_urn) == maybe_urn:
        # already a valid identity in the right form
        return maybe_urn
    elif maybe_urn.startswith("urn:li:"):
        # this looks like an urn, but not a type we recognize
        raise Exception(
            f"Owner urn {maybe_urn} not recognized as one of the supported types (corpuser, corpGroup)"
        )
    else:
        # mint a user urn as the default
        return make_user_urn(maybe_urn)


def _abort_if_non_existent_urn(graph: DataHubGraph, urn: str, operation: str) -> None:
    try:
        parsed_urn: Urn = Urn.from_string(urn)
        entity_type = parsed_urn.get_type()
    except Exception:
        click.secho(f"Provided urn {urn} does not seem valid", fg="red")
        raise click.Abort() from None
    else:
        if not graph.exists(urn):
            click.secho(
                f"{entity_type.title()} {urn} does not exist. Will not {operation}.",
                fg="red",
            )
            raise click.Abort()


def _print_diff(orig_file, new_file):
    with open(orig_file) as fp:
        orig_lines = fp.readlines()
    with open(new_file) as fp:
        new_lines = fp.readlines()

    sys.stdout.writelines(
        difflib.unified_diff(orig_lines, new_lines, orig_file, new_file)
    )


@click.group(cls=DefaultGroup, default="upsert")
def dataproduct() -> None:
    """A group of commands to interact with the DataProduct entity in DataHub."""
    pass


def mutate(file: Path, validate_assets: bool, external_url: str, upsert: bool) -> None:
    """Update or Upsert a Data Product in DataHub"""

    config_dict = load_file(pathlib.Path(file))
    id = config_dict.get("id") if isinstance(config_dict, dict) else None
    with get_default_graph(ClientMode.CLI) as graph:
        data_product: DataProduct = DataProduct.from_yaml(file, graph)
        external_url_override = (
            external_url
            or os.getenv("DATAHUB_DATAPRODUCT_EXTERNAL_URL")
            or data_product.external_url
        )
        data_product.external_url = external_url_override
        if upsert and not graph.exists(data_product.urn):
            logger.info(f"Data Product {data_product.urn} does not exist, will create.")
            upsert = False

        if validate_assets and data_product.assets:
            missing_assets = []
            for asset in data_product.assets:
                try:
                    assert graph.exists(asset)
                except Exception as e:
                    logger.debug("Failed to validate existence", exc_info=e)
                    missing_assets.append(asset)
            if missing_assets:
                for a in missing_assets:
                    click.secho(f"Asset: {a} doesn't exist on DataHub", fg="red")
                click.secho(
                    "Aborting update due to the presence of missing assets in the yaml file. Turn off validation of assets using the --no-validate-assets option if you want to proceed.",
                    fg="red",
                )
                raise click.Abort()
        try:
            for mcp in data_product.generate_mcp(upsert=upsert):
                graph.emit(mcp)
            click.secho(f"Update succeeded for urn {data_product.urn}.", fg="green")
        except Exception as e:
            click.secho(
                f"Update failed for id {id}. due to {e}",
                fg="red",
            )


@dataproduct.command(
    name="update",
)
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@click.option(
    "--validate-assets/--no-validate-assets", required=False, is_flag=True, default=True
)
@click.option("--external-url", required=False, type=str)
@upgrade.check_upgrade
def update(file: Path, validate_assets: bool, external_url: str) -> None:
    """Create or Update a Data Product in DataHub. Use upsert if you want to apply partial updates."""

    mutate(file, validate_assets, external_url, upsert=False)


@dataproduct.command(
    name="upsert",
)
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@click.option(
    "--validate-assets/--no-validate-assets", required=False, is_flag=True, default=True
)
@click.option("--external-url", required=False, type=str)
@upgrade.check_upgrade
def upsert(file: Path, validate_assets: bool, external_url: str) -> None:
    """Upsert attributes to a Data Product in DataHub."""

    mutate(file, validate_assets, external_url, upsert=True)


@dataproduct.command(
    name="diff",
)
@click.option("-f", "--file", required=True, type=click.Path(exists=True))
@click.option("--update", required=False, is_flag=True, default=False)
@upgrade.check_upgrade
def diff(file: Path, update: bool) -> None:
    """Diff a Data Product file with its twin in DataHub"""

    with get_default_graph(ClientMode.CLI) as emitter:
        id: Optional[str] = None
        try:
            data_product_local: DataProduct = DataProduct.from_yaml(file, emitter)
            id = data_product_local.id
            data_product_remote = DataProduct.from_datahub(
                emitter, data_product_local.urn
            )
            with NamedTemporaryFile(suffix=".yaml") as temp_fp:
                update_needed = data_product_remote.patch_yaml(
                    data_product_local,
                    Path(temp_fp.name),
                )
                if not update_needed:
                    click.secho(f"Update not needed for id {id}.", fg="green")
                else:
                    _print_diff(file, temp_fp.name)
                    if update:
                        # copy the temp file over to the main file
                        copyfile(temp_fp.name, file)
                        click.echo(f"Updated {file} successfully.")
                    else:
                        click.secho(f"Update needed for id {id}", fg="red")

        except Exception:
            raise


@dataproduct.command(
    name="delete",
)
@click.option(
    "--urn", required=False, type=str, help="The urn for the data product to delete"
)
@click.option(
    "-f",
    "--file",
    required=False,
    type=click.Path(exists=True),
    help="The file containing the data product definition",
)
@click.option("--hard/--soft", required=False, is_flag=True, default=False)
@upgrade.check_upgrade
def delete(urn: str, file: Path, hard: bool) -> None:
    """Delete a Data Product in DataHub. Defaults to a soft-delete. Use --hard to completely erase metadata."""

    if not urn and not file:
        click.secho(
            "Must provide either an urn or a file to delete a data product", fg="red"
        )
        raise click.Abort()

    graph: DataHubGraph
    with get_default_graph(ClientMode.CLI) as graph:
        data_product_urn = (
            urn if urn.startswith("urn:li:dataProduct") else f"urn:li:dataProduct:{urn}"
        )
        if not urn:
            data_product: DataProduct = DataProduct.from_yaml(file, graph)
            data_product_urn = data_product.urn

        _abort_if_non_existent_urn(graph, data_product_urn, "delete")

        if hard:
            # we only delete references if this is a hard delete
            graph.delete_references_to_urn(data_product_urn)

        graph.delete_entity(data_product_urn, hard=hard)

        click.secho(f"Data Product {data_product_urn} deleted")


@dataproduct.command(
    name="get",
)
@click.option("--urn", required=True, type=str)
@click.option("--to-file", required=False, type=str)
@upgrade.check_upgrade
def get(urn: str, to_file: str) -> None:
    """Get a Data Product from DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"

    with get_default_graph(ClientMode.CLI) as graph:
        if graph.exists(urn):
            dataproduct: DataProduct = DataProduct.from_datahub(graph=graph, id=urn)
            click.secho(
                f"{json.dumps(dataproduct.dict(exclude_unset=True, exclude_none=True), indent=2)}"
            )
            if to_file:
                dataproduct.to_yaml(Path(to_file))
                click.secho(f"Data Product yaml written to {to_file}", fg="green")
        else:
            click.secho(f"Data Product {urn} does not exist")


@dataproduct.command(
    name="set_description",
)
@click.option("--urn", required=True, type=str)
@click.option(
    "--description",
    required=False,
    type=str,
    help="Inline documentation for this data product",
)
@click.option(
    "--md-file",
    required=False,
    type=click.Path(exists=True),
    help="A markdown file that contains documentation for this data product",
)
@upgrade.check_upgrade
def set_description(urn: str, description: str, md_file: Path) -> None:
    """Set description for a Data Product in DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"

    if description is None and not md_file:
        click.secho(
            "Need one of --description or --md-file provided to populate description field",
            fg="red",
        )
        raise click.Abort()

    if description and md_file:
        click.secho(
            "Need only one of --description or --md-file provided to populate description field. You provided both.",
            fg="red",
        )
        raise click.Abort()

    if md_file:
        logger.info(f"Opening file {md_file} for populating description")
        with open(md_file) as fp:
            description = fp.read()

    dataproduct_patcher: DataProductPatchBuilder = DataProduct.get_patch_builder(urn)
    dataproduct_patcher.set_description(description)
    with get_default_graph(ClientMode.CLI) as graph:
        _abort_if_non_existent_urn(graph, urn, "set description")
        for mcp in dataproduct_patcher.build():
            graph.emit(mcp)


@dataproduct.command(name="add_owner", help="Add an owner to a Data Product")
@click.option("--urn", required=True, type=str)
@click.option("--owner", required=True, type=str)
@click.option(
    "--owner-type",
    required=False,
    type=click.Choice(
        [
            OwnershipTypeClass.BUSINESS_OWNER,
            OwnershipTypeClass.TECHNICAL_OWNER,
            OwnershipTypeClass.DATA_STEWARD,
        ],
        case_sensitive=False,
    ),
    default=OwnershipTypeClass.TECHNICAL_OWNER,
)
@upgrade.check_upgrade
def add_owner(urn: str, owner: str, owner_type: str) -> None:
    """Add owner for a Data Product in DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"
    dataproduct_patcher: DataProductPatchBuilder = DataProduct.get_patch_builder(urn)
    owner_type, owner_type_urn = validate_ownership_type(owner_type)
    dataproduct_patcher.add_owner(
        owner=OwnerClass(
            owner=_get_owner_urn(owner), type=owner_type, typeUrn=owner_type_urn
        )
    )
    with get_default_graph(ClientMode.CLI) as graph:
        _abort_if_non_existent_urn(graph, urn, "add owners")
        for mcp in dataproduct_patcher.build():
            graph.emit(mcp)


@dataproduct.command(name="remove_owner", help="Remove an owner from a Data Product")
@click.option("--urn", required=True, type=str)
@click.argument("owner_urn", required=True, type=str)
@upgrade.check_upgrade
def remove_owner(urn: str, owner_urn: str) -> None:
    """Remove owner for a Data Product in DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"
    dataproduct_patcher: DataProductPatchBuilder = DataProduct.get_patch_builder(urn)
    dataproduct_patcher.remove_owner(owner=_get_owner_urn(owner_urn))
    with get_default_graph(ClientMode.CLI) as graph:
        _abort_if_non_existent_urn(graph, urn, "remove owners")
        for mcp in dataproduct_patcher.build():
            click.echo(json.dumps(mcp.to_obj()))
            graph.emit(mcp)


@dataproduct.command(name="add_asset", help="Add an asset to a Data Product")
@click.option("--urn", required=True, type=str)
@click.option("--asset", required=True, type=str)
@click.option(
    "--validate-assets/--no-validate-assets", required=False, is_flag=True, default=True
)
@upgrade.check_upgrade
def add_asset(urn: str, asset: str, validate_assets: bool) -> None:
    """Add asset for a Data Product in DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"
    dataproduct_patcher: DataProductPatchBuilder = DataProduct.get_patch_builder(urn)
    dataproduct_patcher.add_asset(asset)
    with get_default_graph(ClientMode.CLI) as graph:
        _abort_if_non_existent_urn(graph, urn, "add assets")
        if validate_assets:
            _abort_if_non_existent_urn(
                graph,
                asset,
                "add assets. Use --no-validate-assets if you want to turn off validation",
            )
        for mcp in dataproduct_patcher.build():
            graph.emit(mcp)


@dataproduct.command(name="remove_asset", help="Remove an asset from a Data Product")
@click.option("--urn", required=True, type=str)
@click.option("--asset", required=True, type=str)
@click.option(
    "--validate-assets/--no-validate-assets", required=False, is_flag=True, default=True
)
@upgrade.check_upgrade
def remove_asset(urn: str, asset: str, validate_assets: bool) -> None:
    """Remove asset for a Data Product in DataHub"""

    if not urn.startswith("urn:li:dataProduct:"):
        urn = f"urn:li:dataProduct:{urn}"
    dataproduct_patcher: DataProductPatchBuilder = DataProduct.get_patch_builder(urn)
    dataproduct_patcher.remove_asset(asset)
    with get_default_graph(ClientMode.CLI) as graph:
        _abort_if_non_existent_urn(graph, urn, "remove assets")
        if validate_assets:
            _abort_if_non_existent_urn(
                graph,
                asset,
                "remove assets. Use --no-validate-assets if you want to turn off validation",
            )
        for mcp in dataproduct_patcher.build():
            graph.emit(mcp)



import json
import logging
import os
import pathlib
import re
import tempfile
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import pytest

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.sink.file import write_metadata_file
from datahub.metadata.schema_classes import MetadataChangeEventClass
from datahub.metadata.urns import Urn
from datahub.testing.compare_metadata_json import (
    assert_metadata_files_equal,
    load_json_file,
)

logger = logging.getLogger(__name__)

IGNORE_PATH_TIMESTAMPS = [
    # Ignore timestamps from the ETL pipeline. A couple examples:
    r"root\[\d+\]\['proposedSnapshot'\].+\['aspects'\].+\['created'\]\['time'\]",
    r"root\[\d+\]\['proposedSnapshot'\].+\['aspects'\].+\['lastModified'\]\['time'\]",
    r"root\[\d+\]\['proposedSnapshot'\].+\['aspects'\].+\['createStamp'\]\['time'\]",
    r"root\[\d+\]\['proposedSnapshot'\].+\['aspects'\].+\['auditStamp'\]\['time'\]",
]


class MCEConstants:
    PROPOSED_SNAPSHOT = "proposedSnapshot"
    DATASET_SNAPSHOT_CLASS = (
        "com.linkedin.pegasus2avro.metadata.snapshot.DatasetSnapshot"
    )


class MCPConstants:
    CHANGE_TYPE = "changeType"
    ENTITY_URN = "entityUrn"
    ENTITY_TYPE = "entityType"
    ASPECT_NAME = "aspectName"
    ASPECT_VALUE = "aspect"


class EntityType:
    DATASET = "dataset"
    PIPELINE = "dataFlow"
    FLOW = "dataFlow"
    TASK = "dataJob"
    JOB = "dataJob"
    USER = "corpuser"
    GROUP = "corpGroup"


def clean_nones(value):
    """
    Recursively remove all None values from dictionaries and lists, and returns
    the result as a new dictionary or list.
    """
    if isinstance(value, list):
        return [clean_nones(x) for x in value if x is not None]
    elif isinstance(value, dict):
        return {key: clean_nones(val) for key, val in value.items() if val is not None}
    else:
        return value


def check_golden_file(
    pytestconfig: pytest.Config,
    output_path: Union[str, os.PathLike],
    golden_path: Union[str, os.PathLike],
    ignore_paths: Sequence[str] = (),
    ignore_paths_v2: Sequence[str] = (),
    ignore_order: bool = True,
) -> None:
    # TODO: Remove the pytestconfig parameter since it's redundant.
    # Or more straightforward - we can remove the `check_golden_file` method
    # and use assert_metadata_files_equal directly. Maybe call it "check_golden_metadata"?
    # In a lot of cases, the output_path is also just annoying - our pytest setup
    # should be responsible for figuring out where to put the temp file.
    assert_metadata_files_equal(
        output_path=output_path,
        golden_path=golden_path,
        ignore_paths=ignore_paths,
        ignore_paths_v2=ignore_paths_v2,
        ignore_order=ignore_order,
    )


def check_goldens_stream(
    outputs: List,
    golden_path: Union[str, os.PathLike],
    ignore_paths: Sequence[str] = (),
    ignore_order: bool = True,
) -> None:
    with tempfile.NamedTemporaryFile() as f:
        write_metadata_file(pathlib.Path(f.name), outputs)

        assert_metadata_files_equal(
            output_path=f.name,
            golden_path=golden_path,
            ignore_paths=ignore_paths,
            ignore_order=ignore_order,
        )


def _get_field_for_entity_type_in_mce(entity_type: str) -> str:
    """Returns the field to look for depending on the type of entity in the MCE"""
    if entity_type == EntityType.DATASET:
        return MCEConstants.DATASET_SNAPSHOT_CLASS
    raise Exception(f"Not implemented for entity_type {entity_type}")


def _get_filter(
    mce: bool = False, mcp: bool = False, entity_type: Optional[str] = None
) -> Callable[[Dict], bool]:
    if mce:
        # cheap way to determine if we are working with an MCE for the appropriate entity_type
        if entity_type:
            return (
                lambda x: MCEConstants.PROPOSED_SNAPSHOT in x
                and _get_field_for_entity_type_in_mce(str(entity_type))
                in x[MCEConstants.PROPOSED_SNAPSHOT]
            )
        else:
            return lambda x: MCEConstants.PROPOSED_SNAPSHOT in x
    if mcp:
        # cheap way to determine if we are working with an MCP
        return lambda x: MCPConstants.CHANGE_TYPE in x and (
            x[MCPConstants.ENTITY_TYPE] == entity_type if entity_type else True
        )
    return lambda _: False


def _get_element(event: Dict[str, Any], path_spec: List[str]) -> Any:
    try:
        for p in path_spec:
            if p not in event:
                return None
            else:
                event = event.get(p, {})
        return event
    except Exception as e:
        print(event)
        raise e


def _element_matches_pattern(
    event: Dict[str, Any], path_spec: List[str], pattern: str
) -> Tuple[bool, bool]:
    import re

    element = _get_element(event, path_spec)
    if element is None:
        return (False, False)
    else:
        return (True, re.search(pattern, str(element)) is not None)


def get_entity_urns(events_file: str) -> Set[str]:
    events = load_json_file(events_file)
    assert isinstance(events, list)
    return _get_entity_urns(events)


def _get_entity_urns(events_list: List[Dict]) -> Set[str]:
    entity_type = "dataset"
    # mce urns
    mce_urns = {
        _get_element(x, _get_mce_urn_path_spec(entity_type))
        for x in events_list
        if _get_filter(mce=True, entity_type=entity_type)(x)
    }
    mcp_urns = {
        _get_element(x, _get_mcp_urn_path_spec())
        for x in events_list
        if _get_filter(mcp=True, entity_type=entity_type)(x)
    }
    all_urns = mce_urns.union(mcp_urns)
    return all_urns


def assert_mcp_entity_urn(
    filter: str, entity_type: str, regex_pattern: str, file: str
) -> int:
    def get_path_spec_for_urn() -> List[str]:
        return [MCPConstants.ENTITY_URN]

    test_output = load_json_file(file)
    if isinstance(test_output, list):
        path_spec = get_path_spec_for_urn()
        filter_operator = _get_filter(mcp=True, entity_type=entity_type)
        filtered_events = [
            (x, _element_matches_pattern(x, path_spec, regex_pattern))
            for x in test_output
            if filter_operator(x)
        ]
        failed_events = [y for y in filtered_events if not y[1][0] or not y[1][1]]
        if failed_events:
            raise Exception("Failed to match events", failed_events)
        return len(filtered_events)
    else:
        raise Exception(
            f"Did not expect the file {file} to not contain a list of items"
        )


def _get_mce_urn_path_spec(entity_type: str) -> List[str]:
    if entity_type == EntityType.DATASET:
        return [
            MCEConstants.PROPOSED_SNAPSHOT,
            MCEConstants.DATASET_SNAPSHOT_CLASS,
            "urn",
        ]
    raise Exception(f"Not implemented for entity_type: {entity_type}")


def _get_mcp_urn_path_spec() -> List[str]:
    return [MCPConstants.ENTITY_URN]


def assert_mce_entity_urn(
    filter: str, entity_type: str, regex_pattern: str, file: str
) -> int:
    """Assert that all mce entity urns must match the regex pattern passed in. Return the number of events matched"""

    test_output = load_json_file(file)
    if isinstance(test_output, list):
        path_spec = _get_mce_urn_path_spec(entity_type)
        filter_operator = _get_filter(mce=True)
        filtered_events = [
            (x, _element_matches_pattern(x, path_spec, regex_pattern))
            for x in test_output
            if filter_operator(x)
        ]
        failed_events = [y for y in filtered_events if not y[1][0] or not y[1][1]]
        if failed_events:
            raise Exception(
                "Failed to match events: {json.dumps(failed_events, indent=2)}"
            )
        return len(filtered_events)
    else:
        raise Exception(
            f"Did not expect the file {file} to not contain a list of items"
        )


def assert_for_each_entity(
    entity_type: str,
    aspect_name: str,
    aspect_field_matcher: Dict[str, Any],
    file: str,
    exception_urns: Optional[List[str]] = None,
) -> int:
    """Assert that an aspect name with the desired fields exists for each entity urn"""
    if exception_urns is None:
        exception_urns = []
    test_output = load_json_file(file)
    assert isinstance(test_output, list)
    # mce urns
    mce_urns = {
        _get_element(x, _get_mce_urn_path_spec(entity_type))
        for x in test_output
        if _get_filter(mce=True, entity_type=entity_type)(x)
    }
    mcp_urns = {
        _get_element(x, _get_mcp_urn_path_spec())
        for x in test_output
        if _get_filter(mcp=True, entity_type=entity_type)(x)
    }
    all_urns = mce_urns.union(mcp_urns)
    # there should not be any None urns
    assert None not in all_urns
    aspect_map = {urn: None for urn in all_urns}
    # iterate over all mcps
    for o in [
        mcp
        for mcp in test_output
        if _get_filter(mcp=True, entity_type=entity_type)(mcp)
    ]:
        if o.get(MCPConstants.ASPECT_NAME) == aspect_name:
            # load the inner aspect payload and assign to this urn
            aspect_map[o[MCPConstants.ENTITY_URN]] = o.get(
                MCPConstants.ASPECT_VALUE, {}
            ).get("json")

    success: List[str] = []
    failures: List[str] = []
    for urn, aspect_val in aspect_map.items():
        if aspect_val is not None:
            for f in aspect_field_matcher:
                assert aspect_field_matcher[f] == _get_element(aspect_val, [f]), (
                    f"urn: {urn} -> Field {f} must match value {aspect_field_matcher[f]}, found {_get_element(aspect_val, [f])}"
                )
            success.append(urn)
        elif urn not in exception_urns:
            print(f"Adding {urn} to failures")
            failures.append(urn)

    if success:
        print(f"Succeeded on assertion for urns {success}")
    if failures:
        raise AssertionError(
            f"Failed to find aspect_name {aspect_name} for urns {json.dumps(failures, indent=2)}"
        )

    return len(success)


def assert_entity_mce_aspect(
    entity_urn: str, aspect: Any, aspect_type: Type, file: str
) -> int:
    # TODO: Replace with read_metadata_file()
    test_output = load_json_file(file)
    entity_type = Urn.from_string(entity_urn).entity_type
    assert isinstance(test_output, list)
    # mce urns
    mces: List[MetadataChangeEventClass] = [
        MetadataChangeEventClass.from_obj(x)
        for x in test_output
        if _get_filter(mce=True, entity_type=entity_type)(x)
        and _get_element(x, _get_mce_urn_path_spec(entity_type)) == entity_urn
    ]
    matches = 0
    for mce in mces:
        for a in mce.proposedSnapshot.aspects:
            if isinstance(a, aspect_type):
                assert a == aspect
                matches = matches + 1
    return matches


def assert_entity_mcp_aspect(
    entity_urn: str, aspect_field_matcher: Dict[str, Any], aspect_name: str, file: str
) -> int:
    # TODO: Replace with read_metadata_file()
    test_output = load_json_file(file)
    entity_type = Urn.from_string(entity_urn).entity_type
    assert isinstance(test_output, list)
    # mcps that match entity_urn
    mcps: List[MetadataChangeProposalWrapper] = [
        MetadataChangeProposalWrapper.from_obj_require_wrapper(x)
        for x in test_output
        if _get_filter(mcp=True, entity_type=entity_type)(x)
        and _get_element(x, _get_mcp_urn_path_spec()) == entity_urn
    ]
    matches = 0
    for mcp in mcps:
        if mcp.aspectName == aspect_name:
            assert mcp.aspect
            aspect_val = mcp.aspect.to_obj()
            for f in aspect_field_matcher:
                assert aspect_field_matcher[f] == _get_element(aspect_val, [f]), (
                    f"urn: {mcp.entityUrn} -> Field {f} must match value {aspect_field_matcher[f]}, found {_get_element(aspect_val, [f])}"
                )
                matches = matches + 1
    return matches


def assert_entity_urn_not_like(entity_type: str, regex_pattern: str, file: str) -> int:
    """Assert that there are no entity urns that match the regex pattern passed in. Returns the total number of events in the file"""

    # TODO: Refactor common code with assert_entity_urn_like.
    test_output = load_json_file(file)
    assert isinstance(test_output, list)
    # mce urns
    mce_urns = {
        _get_element(x, _get_mce_urn_path_spec(entity_type))
        for x in test_output
        if _get_filter(mce=True, entity_type=entity_type)(x)
    }
    mcp_urns = {
        _get_element(x, _get_mcp_urn_path_spec())
        for x in test_output
        if _get_filter(mcp=True, entity_type=entity_type)(x)
    }
    all_urns = mce_urns.union(mcp_urns)
    print(all_urns)
    matched_urns = [u for u in all_urns if re.match(regex_pattern, u)]
    if matched_urns:
        raise AssertionError(f"urns found that match the deny list {matched_urns}")
    return len(test_output)


def assert_entity_urn_like(entity_type: str, regex_pattern: str, file: str) -> int:
    """Assert that there exist entity urns that match the regex pattern passed in. Returns the total number of events in the file"""

    test_output = load_json_file(file)
    assert isinstance(test_output, list)
    # mce urns
    mce_urns = {
        _get_element(x, _get_mce_urn_path_spec(entity_type))
        for x in test_output
        if _get_filter(mce=True, entity_type=entity_type)(x)
    }
    mcp_urns = {
        _get_element(x, _get_mcp_urn_path_spec())
        for x in test_output
        if _get_filter(mcp=True, entity_type=entity_type)(x)
    }
    all_urns = mce_urns.union(mcp_urns)
    print(all_urns)
    matched_urns = [u for u in all_urns if re.match(regex_pattern, u)]
    if matched_urns:
        return len(matched_urns)
    else:
        raise AssertionError(
            f"No urns found that match the pattern {regex_pattern}. Full list is {all_urns}"
        )



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from datahub.emitter.mce_builder import (
    make_data_flow_urn,
    make_data_job_urn,
    make_data_platform_urn,
    make_dataplatform_instance_urn,
)
from datahub.emitter.mcp_builder import (
    DatabaseKey,
    SchemaKey,
)
from datahub.ingestion.source.common.subtypes import (
    FlowContainerSubTypes,
    JobContainerSubTypes,
)
from datahub.ingestion.source.sql.stored_procedures.base import BaseProcedure
from datahub.metadata.schema_classes import (
    ContainerClass,
    DataFlowInfoClass,
    DataJobInfoClass,
    DataJobInputOutputClass,
    DataPlatformInstanceClass,
    SubTypesClass,
)


@dataclass
class ProcedureDependency:
    db: str
    schema: str
    name: str
    type: str
    env: str
    server: Optional[str]
    source: str = "mssql"


@dataclass
class ProcedureLineageStream:
    dependencies: List[ProcedureDependency]

    @property
    def as_property(self) -> Dict[str, str]:
        return {
            f"{dep.db}.{dep.schema}.{dep.name}": dep.type for dep in self.dependencies
        }


@dataclass
class MSSQLJob:
    db: str
    platform_instance: Optional[str]
    name: str
    env: str
    source: str = "mssql"
    type: str = "JOB"

    @property
    def formatted_name(self) -> str:
        return self.name.replace(",", "-")

    @property
    def full_type(self) -> str:
        return f"({self.source},{self.formatted_name},{self.env})"

    @property
    def orchestrator(self) -> str:
        return self.source

    @property
    def cluster(self) -> str:
        return f"{self.env}"


@dataclass
class MSSQLProceduresContainer:
    db: str
    platform_instance: Optional[str]
    name: str
    env: str
    source: str = "mssql"
    type: str = "JOB"

    @property
    def formatted_name(self) -> str:
        return self.name.replace(",", "-")

    @property
    def orchestrator(self) -> str:
        return self.source

    @property
    def cluster(self) -> str:
        return f"{self.env}"

    @property
    def full_type(self) -> str:
        return f"({self.source},{self.name},{self.env})"


@dataclass
class ProcedureParameter:
    name: str
    type: str

    @property
    def properties(self) -> Dict[str, str]:
        return {"type": self.type}


@dataclass
class StoredProcedure:
    db: str
    schema: str
    name: str
    flow: Union[MSSQLJob, MSSQLProceduresContainer]
    type: str = "STORED_PROCEDURE"
    source: str = "mssql"
    code: Optional[str] = None

    @property
    def full_type(self) -> str:
        return self.source.upper() + "_" + self.type

    @property
    def formatted_name(self) -> str:
        return self.name.replace(",", "-")

    @property
    def full_name(self) -> str:
        return f"{self.db}.{self.schema}.{self.formatted_name}"

    @property
    def escape_full_name(self) -> str:
        return f"[{self.db}].[{self.schema}].[{self.formatted_name}]"

    def to_base_procedure(self) -> BaseProcedure:
        return BaseProcedure(
            name=self.formatted_name,
            procedure_definition=self.code,
            created=None,
            last_altered=None,
            comment=None,
            argument_signature=None,
            return_type=None,
            language="SQL",
            extra_properties=None,
        )


@dataclass
class JobStep:
    job_name: str
    step_name: str
    flow: MSSQLJob
    type: str = "JOB_STEP"
    source: str = "mssql"

    @property
    def formatted_step(self) -> str:
        return self.step_name.replace(",", "-").replace(" ", "_").lower()

    @property
    def formatted_name(self) -> str:
        return self.job_name.replace(",", "-")

    @property
    def full_type(self) -> str:
        return self.source.upper() + "_" + self.type

    @property
    def full_name(self) -> str:
        return self.formatted_name


@dataclass
class MSSQLDataJob:
    entity: Union[StoredProcedure, JobStep]
    type: str = "dataJob"
    source: str = "mssql"
    external_url: Optional[str] = None
    description: Optional[str] = None
    status: Optional[str] = None
    incoming: List[str] = field(default_factory=list)
    outgoing: List[str] = field(default_factory=list)
    input_jobs: List[str] = field(default_factory=list)
    job_properties: Dict[str, str] = field(default_factory=dict)

    @property
    def urn(self) -> str:
        return make_data_job_urn(
            orchestrator=self.entity.flow.orchestrator,
            flow_id=self.entity.flow.formatted_name,
            job_id=self.entity.formatted_name,
            cluster=self.entity.flow.cluster,
            platform_instance=self.entity.flow.platform_instance,
        )

    def add_property(
        self,
        name: str,
        value: str,
    ) -> None:
        self.job_properties[name] = value

    @property
    def valued_properties(self) -> Dict[str, str]:
        if self.job_properties:
            return {k: v for k, v in self.job_properties.items() if v is not None}
        return self.job_properties

    @property
    def as_datajob_input_output_aspect(self) -> DataJobInputOutputClass:
        return DataJobInputOutputClass(
            inputDatasets=sorted(self.incoming),
            outputDatasets=sorted(self.outgoing),
            inputDatajobs=sorted(self.input_jobs),
        )

    @property
    def as_datajob_info_aspect(self) -> DataJobInfoClass:
        return DataJobInfoClass(
            name=self.entity.full_name,
            type=self.entity.full_type,
            description=self.description,
            customProperties=self.valued_properties,
            externalUrl=self.external_url,
            status=self.status,
        )

    @property
    def as_subtypes_aspect(self) -> SubTypesClass:
        assert isinstance(self.entity, (JobStep, StoredProcedure))
        type = (
            JobContainerSubTypes.MSSQL_JOBSTEP
            if isinstance(self.entity, JobStep)
            else JobContainerSubTypes.STORED_PROCEDURE
        )
        return SubTypesClass(
            typeNames=[type],
        )

    @property
    def as_maybe_platform_instance_aspect(self) -> Optional[DataPlatformInstanceClass]:
        if self.entity.flow.platform_instance:
            return DataPlatformInstanceClass(
                platform=make_data_platform_urn(self.entity.flow.orchestrator),
                instance=make_dataplatform_instance_urn(
                    platform=self.entity.flow.orchestrator,
                    instance=self.entity.flow.platform_instance,
                ),
            )
        return None

    @property
    def as_container_aspect(self) -> ContainerClass:
        key_args = dict(
            platform=self.entity.flow.orchestrator,
            instance=self.entity.flow.platform_instance,
            env=self.entity.flow.env,
            database=self.entity.flow.db,
        )
        container_key = (
            SchemaKey(
                schema=self.entity.schema,
                **key_args,
            )
            if isinstance(self.entity, StoredProcedure)
            else DatabaseKey(
                **key_args,
            )
        )
        return ContainerClass(container=container_key.as_urn())


@dataclass
class MSSQLDataFlow:
    entity: Union[MSSQLJob, MSSQLProceduresContainer]
    type: str = "dataFlow"
    source: str = "mssql"
    external_url: Optional[str] = None
    flow_properties: Dict[str, str] = field(default_factory=dict)

    def add_property(
        self,
        name: str,
        value: str,
    ) -> None:
        self.flow_properties[name] = value

    @property
    def urn(self) -> str:
        return make_data_flow_urn(
            orchestrator=self.entity.orchestrator,
            flow_id=self.entity.formatted_name,
            cluster=self.entity.cluster,
            platform_instance=self.entity.platform_instance,
        )

    @property
    def as_dataflow_info_aspect(self) -> DataFlowInfoClass:
        return DataFlowInfoClass(
            name=self.entity.formatted_name,
            customProperties=self.flow_properties,
            externalUrl=self.external_url,
        )

    @property
    def as_subtypes_aspect(self) -> SubTypesClass:
        assert isinstance(self.entity, (MSSQLJob, MSSQLProceduresContainer))
        type = (
            FlowContainerSubTypes.MSSQL_JOB
            if isinstance(self.entity, MSSQLJob)
            else FlowContainerSubTypes.MSSQL_PROCEDURE_CONTAINER
        )
        return SubTypesClass(
            typeNames=[type],
        )

    @property
    def as_maybe_platform_instance_aspect(self) -> Optional[DataPlatformInstanceClass]:
        if self.entity.platform_instance:
            return DataPlatformInstanceClass(
                platform=make_data_platform_urn(self.entity.orchestrator),
                instance=make_dataplatform_instance_urn(
                    self.entity.orchestrator, self.entity.platform_instance
                ),
            )
        return None

    @property
    def as_container_aspect(self) -> ContainerClass:
        databaseKey = DatabaseKey(
            platform=self.entity.orchestrator,
            instance=self.entity.platform_instance,
            env=self.entity.env,
            database=self.entity.db,
        )
        return ContainerClass(container=databaseKey.as_urn())



import logging
import socket
import time
from typing import Iterable, Optional
from urllib.parse import urlparse

import dns.exception
import dns.resolver
import requests

from datahub.configuration.common import ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit

logger = logging.getLogger(__name__)


class DataHubDebugSourceConfig(ConfigModel):
    dns_probe_url: Optional[str] = None


@platform_name("DataHubDebug")
@config_class(DataHubDebugSourceConfig)
@support_status(SupportStatus.TESTING)
class DataHubDebugSource(Source):
    """
    DataHubDebugSource is helper to debug things in executor where ingestion is running.

    This source can perform the following tasks:
    1. Network probe of a URL. Different from test connection in sources as that is after source starts.

    """

    def __init__(self, ctx: PipelineContext, config: DataHubDebugSourceConfig):
        self.ctx = ctx
        self.config = config
        self.report = SourceReport()
        self.report.event_not_produced_warn = False

    @classmethod
    def create(cls, config_dict, ctx):
        config = DataHubDebugSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def perform_dns_probe(self, url: str) -> None:
        """
        Perform comprehensive DNS probe and network connectivity tests.
        Logs detailed information to help diagnose network issues.
        """
        logger.info(f"Starting DNS probe for URL: {url}")
        logger.info("=" * 60)
        logger.info(f"DNS PROBE REPORT FOR: {url}")
        logger.info("=" * 60)

        try:
            # Parse the URL to extract hostname
            parsed_url = urlparse(
                url if url.startswith(("http://", "https://")) else f"http://{url}"
            )
            hostname = parsed_url.hostname or parsed_url.netloc
            port = parsed_url.port or (443 if parsed_url.scheme == "https" else 80)

            logger.info(f"Parsed hostname: {hostname}")
            logger.info(f"Target port: {port}")
            logger.info(f"URL scheme: {parsed_url.scheme}")
            logger.info("-" * 60)

            # Test 1: Enhanced DNS resolution with dnspython if available
            logger.info("1. DNS RESOLUTION TEST")
            self._dns_probe_with_dnspython(hostname)

            logger.info("-" * 60)

            # Test 2: HTTP/HTTPS connectivity test with requests if available
            logger.info("2. HTTP CONNECTIVITY TEST")
            self._http_probe_with_requests(url)

            logger.info("-" * 60)

            # Test 3: System network information
            logger.info("3. SYSTEM NETWORK INFORMATION")
            self._log_system_network_info()

        except Exception as e:
            logger.error(f"DNS probe failed with unexpected error: {e}", exc_info=True)

        logger.info("=" * 60)
        logger.info("DNS PROBE COMPLETED")
        logger.info("=" * 60)

    def _dns_probe_with_dnspython(self, hostname: str) -> None:
        """Enhanced DNS probing using dnspython library"""
        try:
            # Test different record types
            record_types = ["A", "AAAA", "CNAME", "MX"]

            for record_type in record_types:
                try:
                    start_time = time.time()
                    answers = dns.resolver.resolve(hostname, record_type)
                    dns_time = time.time() - start_time

                    logger.info(
                        f"✓ {record_type} record resolution successful ({dns_time:.3f}s)"
                    )
                    for answer in answers:
                        logger.info(f"  - {record_type}: {answer}")

                except dns.resolver.NXDOMAIN:
                    logger.info(f"✗ {record_type} record: Domain does not exist")
                except dns.resolver.NoAnswer:
                    logger.info(
                        f"- {record_type} record: No answer (record type not available)"
                    )
                except dns.exception.Timeout:
                    logger.error(f"✗ {record_type} record: DNS query timed out")
                except Exception as e:
                    logger.error(f"✗ {record_type} record query failed: {e}")

            # Test different DNS servers
            logger.info("Testing with different DNS servers:")
            dns_servers = ["8.8.8.8", "1.1.1.1", "208.67.222.222"]

            for dns_server in dns_servers:
                try:
                    resolver = dns.resolver.Resolver()
                    resolver.nameservers = [dns_server]
                    resolver.timeout = 5

                    start_time = time.time()
                    answers = resolver.resolve(hostname, "A")
                    dns_time = time.time() - start_time

                    logger.info(
                        f"✓ DNS server {dns_server} responded ({dns_time:.3f}s)"
                    )
                    for answer in answers:
                        logger.info(f"  - A: {answer}")

                except Exception as e:
                    logger.error(f"✗ DNS server {dns_server} failed: {e}")

        except Exception as e:
            logger.error(f"Enhanced DNS probe failed: {e}", exc_info=True)

    def _http_probe_with_requests(self, url: str) -> None:
        """HTTP connectivity test using requests library"""
        try:
            # Test with different timeouts and methods
            timeout = 10
            allow_redirects_head = True
            allow_redirects_get = False

            # Test HEAD request
            try:
                logger.info(f"Testing HEAD request with timeout {timeout}s")
                start_time = time.time()

                response = requests.head(
                    url, timeout=timeout, allow_redirects=allow_redirects_head
                )

                request_time = time.time() - start_time

                logger.info(f"✓ HEAD request successful ({request_time:.3f}s)")
                logger.info(f"  Status code: {response.status_code}")
                logger.info(
                    f"  Response headers: {dict(list(response.headers.items())[:5])}"
                )

                if hasattr(response, "url") and response.url != url:
                    logger.info(f"  Final URL after redirects: {response.url}")

            except requests.exceptions.Timeout:
                logger.error(f"✗ HEAD request timed out after {timeout}s")
            except requests.exceptions.ConnectionError as e:
                logger.error(f"✗ HEAD connection error: {e}")
            except requests.exceptions.RequestException as e:
                logger.error(f"✗ HEAD request failed: {e}")
            except Exception as e:
                logger.error(f"✗ HEAD unexpected error: {e}")

            # Test GET request
            try:
                logger.info(f"Testing GET request with timeout {timeout}s")
                start_time = time.time()

                response = requests.get(
                    url, timeout=timeout, allow_redirects=allow_redirects_get
                )

                request_time = time.time() - start_time

                logger.info(f"✓ GET request successful ({request_time:.3f}s)")
                logger.info(f"  Status code: {response.status_code}")
                logger.info(
                    f"  Response headers: {dict(list(response.headers.items())[:5])}"
                )

                if hasattr(response, "url") and response.url != url:
                    logger.info(f"  Final URL after redirects: {response.url}")

            except requests.exceptions.Timeout:
                logger.error(f"✗ GET request timed out after {timeout}s")
            except requests.exceptions.ConnectionError as e:
                logger.error(f"✗ GET connection error: {e}")
            except requests.exceptions.RequestException as e:
                logger.error(f"✗ GET request failed: {e}")
            except Exception as e:
                logger.error(f"✗ GET unexpected error: {e}")

        except Exception as e:
            logger.error(f"HTTP probe failed: {e}", exc_info=True)

    def _log_dns_troubleshooting(self) -> None:
        """Log DNS troubleshooting information"""
        logger.info("DNS TROUBLESHOOTING SUGGESTIONS:")
        logger.info("- Check if the hostname is correct")
        logger.info("- Verify DNS server configuration")
        logger.info("- Check network connectivity")
        logger.info("- Try using a different DNS server (8.8.8.8, 1.1.1.1)")
        logger.info("- Check if there are firewall restrictions")

    def _log_system_network_info(self) -> None:
        """Log system network configuration information"""
        try:
            local_hostname = socket.gethostname()
            logger.info(f"Local hostname: {local_hostname}")

            try:
                local_ips = socket.getaddrinfo(local_hostname, None)
                logger.info("Local IP addresses:")
                for addr_info in local_ips:
                    if addr_info[0] in [socket.AF_INET, socket.AF_INET6]:
                        family = "IPv4" if addr_info[0] == socket.AF_INET else "IPv6"
                        logger.info(f"  - {addr_info[4][0]} ({family})")
            except Exception as e:
                logger.warning(f"Could not retrieve local IP addresses: {e}")

            logger.info("DNS Server Connectivity:")
            dns_servers = ["8.8.8.8", "1.1.1.1", "208.67.222.222"]
            for dns_server in dns_servers:
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    sock.settimeout(5)
                    result = sock.connect_ex((dns_server, 53))
                    if result == 0:
                        logger.info(f"  ✓ Can reach {dns_server}:53")
                    else:
                        logger.error(f"  ✗ Cannot reach {dns_server}:53")
                    sock.close()
                except Exception as e:
                    logger.error(f"  ✗ Error testing {dns_server}:53 - {e}")

        except Exception as e:
            logger.warning(f"Could not gather system network info: {e}")

    def _test_alternative_dns(self, hostname: str) -> None:
        """Test hostname resolution using alternative methods"""
        try:
            families = [(socket.AF_INET, "IPv4"), (socket.AF_INET6, "IPv6")]

            for family, family_name in families:
                try:
                    result = socket.getaddrinfo(hostname, None, family)
                    if result:
                        logger.info(f"✓ {family_name} resolution successful:")
                        for addr_info in result[:3]:
                            logger.info(f"  - {addr_info[4][0]}")
                    else:
                        logger.warning(
                            f"✗ {family_name} resolution returned no results"
                        )
                except socket.gaierror:
                    logger.error(f"✗ {family_name} resolution failed")
                except Exception as e:
                    logger.error(f"✗ {family_name} resolution error: {e}")

        except Exception as e:
            logger.error(f"Alternative DNS test failed: {e}")

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        if self.config.dns_probe_url is not None:
            # Perform DNS probe
            logger.info(f"Performing DNS probe for: {self.config.dns_probe_url}")
            self.perform_dns_probe(self.config.dns_probe_url)

        yield from []

    def get_report(self) -> SourceReport:
        return self.report



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterable, List, Optional

from pydantic import Field, SecretStr, root_validator
from typing_extensions import assert_never

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.datetimes import parse_user_datetime
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import SourceCapabilityModifier
from datahub.ingestion.source.hex.api import HexApi, HexApiReport
from datahub.ingestion.source.hex.constants import (
    DATAHUB_API_PAGE_SIZE_DEFAULT,
    HEX_API_BASE_URL_DEFAULT,
    HEX_API_PAGE_SIZE_DEFAULT,
    HEX_PLATFORM_NAME,
)
from datahub.ingestion.source.hex.mapper import Mapper
from datahub.ingestion.source.hex.model import Component, Project
from datahub.ingestion.source.hex.query_fetcher import (
    HexQueryFetcher,
    HexQueryFetcherReport,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.sdk.main_client import DataHubClient


class HexSourceConfig(
    StatefulIngestionConfigBase, PlatformInstanceConfigMixin, EnvConfigMixin
):
    workspace_name: str = Field(
        description="Hex workspace name. You can find this name in your Hex home page URL: https://app.hex.tech/<workspace_name>",
    )
    token: SecretStr = Field(
        description="Hex API token; either PAT or Workflow token - https://learn.hex.tech/docs/api/api-overview#authentication",
    )
    base_url: str = Field(
        default=HEX_API_BASE_URL_DEFAULT,
        description="Hex API base URL. For most Hex users, this will be https://app.hex.tech/api/v1. "
        "Single-tenant app users should replace this with the URL they use to access Hex.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None,
        description="Configuration for stateful ingestion and stale metadata removal.",
    )
    include_components: bool = Field(
        default=True,
        desciption="Include Hex Components in the ingestion",
    )
    page_size: int = Field(
        default=HEX_API_PAGE_SIZE_DEFAULT,
        description="Number of items to fetch per Hex API call.",
    )
    patch_metadata: bool = Field(
        default=False,
        description="Emit metadata as patch events",
    )
    collections_as_tags: bool = Field(
        default=True,
        description="Emit Hex Collections as tags",
    )
    status_as_tag: bool = Field(
        default=True,
        description="Emit Hex Status as tags",
    )
    categories_as_tags: bool = Field(
        default=True,
        description="Emit Hex Category as tags",
    )
    project_title_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex pattern for project titles to filter in ingestion.",
    )
    component_title_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex pattern for component titles to filter in ingestion.",
    )
    set_ownership_from_email: bool = Field(
        default=True,
        description="Set ownership identity from owner/creator email",
    )
    include_lineage: bool = Field(
        default=True,
        description='Include Hex lineage, being fetched from DataHub. See "Limitations" section in the docs for more details about the limitations of this feature.',
    )
    lineage_start_time: Optional[datetime] = Field(
        default=None,
        description="Earliest date of lineage to consider. Default: 1 day before lineage end time. You can specify absolute time like '2023-01-01' or relative time like '-7 days' or '-7d'.",
    )
    lineage_end_time: Optional[datetime] = Field(
        default=None,
        description="Latest date of lineage to consider. Default: Current time in UTC. You can specify absolute time like '2023-01-01' or relative time like '-1 day' or '-1d'.",
    )
    datahub_page_size: int = Field(
        default=DATAHUB_API_PAGE_SIZE_DEFAULT,
        description="Number of items to fetch per DataHub API call.",
    )

    @root_validator(pre=True)
    def validate_lineage_times(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        # lineage_end_time default = now
        if "lineage_end_time" not in data or data["lineage_end_time"] is None:
            data["lineage_end_time"] = datetime.now(tz=timezone.utc)
        # if string is given, parse it
        if isinstance(data["lineage_end_time"], str):
            data["lineage_end_time"] = parse_user_datetime(data["lineage_end_time"])
        # if no timezone is given, assume UTC
        if data["lineage_end_time"].tzinfo is None:
            data["lineage_end_time"] = data["lineage_end_time"].replace(
                tzinfo=timezone.utc
            )
        # at this point, we ensure there is a non null datetime with UTC timezone for lineage_end_time
        assert (
            data["lineage_end_time"]
            and isinstance(data["lineage_end_time"], datetime)
            and data["lineage_end_time"].tzinfo is not None
            and data["lineage_end_time"].tzinfo == timezone.utc
        )

        # lineage_start_time default = lineage_end_time - 1 day
        if "lineage_start_time" not in data or data["lineage_start_time"] is None:
            data["lineage_start_time"] = data["lineage_end_time"] - timedelta(days=1)
        # if string is given, parse it
        if isinstance(data["lineage_start_time"], str):
            data["lineage_start_time"] = parse_user_datetime(data["lineage_start_time"])
        # if no timezone is given, assume UTC
        if data["lineage_start_time"].tzinfo is None:
            data["lineage_start_time"] = data["lineage_start_time"].replace(
                tzinfo=timezone.utc
            )
        # at this point, we ensure there is a non null datetime with UTC timezone for lineage_start_time
        assert (
            data["lineage_start_time"]
            and isinstance(data["lineage_start_time"], datetime)
            and data["lineage_start_time"].tzinfo is not None
            and data["lineage_start_time"].tzinfo == timezone.utc
        )

        return data


@dataclass
class HexReport(
    StaleEntityRemovalSourceReport,
    HexApiReport,
    IngestionStageReport,
    HexQueryFetcherReport,
):
    pass


@platform_name("Hex")
@config_class(HexSourceConfig)
@support_status(SupportStatus.TESTING)
@capability(SourceCapability.DESCRIPTIONS, "Supported by default")
@capability(SourceCapability.OWNERSHIP, "Supported by default")
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.CONTAINERS, "Enabled by default")
@capability(
    SourceCapability.USAGE_STATS,
    "Supported by default",
    subtype_modifier=[
        SourceCapabilityModifier.HEX_PROJECT,
    ],
)
class HexSource(StatefulIngestionSourceBase):
    def __init__(self, config: HexSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.source_config = config
        self.report: HexReport = HexReport()
        self.platform = HEX_PLATFORM_NAME
        self.hex_api = HexApi(
            report=self.report,
            token=self.source_config.token.get_secret_value(),
            base_url=self.source_config.base_url,
            page_size=self.source_config.page_size,
        )
        self.mapper = Mapper(
            workspace_name=self.source_config.workspace_name,
            platform_instance=self.source_config.platform_instance,
            env=self.source_config.env,
            base_url=self.source_config.base_url,
            patch_metadata=self.source_config.patch_metadata,
            collections_as_tags=self.source_config.collections_as_tags,
            status_as_tag=self.source_config.status_as_tag,
            categories_as_tags=self.source_config.categories_as_tags,
            set_ownership_from_email=self.source_config.set_ownership_from_email,
        )
        self.project_registry: Dict[str, Project] = {}
        self.component_registry: Dict[str, Component] = {}

        self.datahub_client: Optional[DataHubClient] = None
        self.query_fetcher: Optional[HexQueryFetcher] = None
        if self.source_config.include_lineage:
            graph = ctx.require_graph("Lineage")
            assert self.source_config.lineage_start_time and isinstance(
                self.source_config.lineage_start_time, datetime
            )
            assert self.source_config.lineage_end_time and isinstance(
                self.source_config.lineage_end_time, datetime
            )
            self.datahub_client = DataHubClient(graph=graph)
            self.query_fetcher = HexQueryFetcher(
                datahub_client=self.datahub_client,
                workspace_name=self.source_config.workspace_name,
                start_datetime=self.source_config.lineage_start_time,
                end_datetime=self.source_config.lineage_end_time,
                report=self.report,
                page_size=self.source_config.datahub_page_size,
            )

    @classmethod
    def create(cls, config_dict: Dict[str, Any], ctx: PipelineContext) -> "HexSource":
        config = HexSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.source_config, self.ctx
            ).workunit_processor,
        ]

    def get_report(self) -> HexReport:
        return self.report

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        with self.report.new_stage("Fetch Hex assets from Hex API"):
            for project_or_component in self.hex_api.fetch_projects():
                if isinstance(project_or_component, Project):
                    if self.source_config.project_title_pattern.allowed(
                        project_or_component.title
                    ):
                        self.project_registry[project_or_component.id] = (
                            project_or_component
                        )
                elif isinstance(project_or_component, Component):
                    if (
                        self.source_config.include_components
                        and self.source_config.component_title_pattern.allowed(
                            project_or_component.title
                        )
                    ):
                        self.component_registry[project_or_component.id] = (
                            project_or_component
                        )
                else:
                    assert_never(project_or_component)

        if self.source_config.include_lineage:
            assert self.datahub_client and self.query_fetcher

            with self.report.new_stage(
                "Fetch Hex lineage from existing Queries in DataHub"
            ):
                for query_metadata in self.query_fetcher.fetch():
                    project = self.project_registry.get(query_metadata.hex_project_id)
                    if project:
                        project.upstream_datasets.extend(
                            query_metadata.dataset_subjects
                        )
                        project.upstream_schema_fields.extend(
                            query_metadata.schema_field_subjects
                        )
                    else:
                        self.report.report_warning(
                            title="Missing project for lineage",
                            message="Lineage missed because missed project, likely due to filter patterns or deleted project.",
                            context=str(
                                dict(
                                    query_urn=query_metadata.urn,
                                    hex_project_id=query_metadata.hex_project_id,
                                )
                            ),
                        )

        with self.report.new_stage("Emit"):
            yield from self.mapper.map_workspace()

            for project in self.project_registry.values():
                yield from self.mapper.map_project(project=project)
            for component in self.component_registry.values():
                yield from self.mapper.map_component(component=component)



import json
import logging
import time
from datetime import datetime
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar

from sqlalchemy import create_engine, text

from datahub.emitter.aspect import ASPECT_MAP
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.serialization_helper import post_json_transform
from datahub.ingestion.source.datahub.config import DataHubSourceConfig
from datahub.ingestion.source.datahub.report import DataHubSourceReport
from datahub.ingestion.source.sql.sql_config import SQLAlchemyConnectionConfig
from datahub.metadata.schema_classes import SystemMetadataClass
from datahub.utilities.lossy_collections import LossyDict, LossyList

logger = logging.getLogger(__name__)

# Should work for at least mysql, mariadb, postgres
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
DATE_FORMAT = "%Y-%m-%d"

ROW = TypeVar("ROW", bound=Dict[str, Any])


class VersionOrderer(Generic[ROW]):
    """Orders rows by (createdon, version == 0).

    That is, orders rows first by createdon, and for equal timestamps, puts version 0 rows last.
    """

    def __init__(self, enabled: bool):
        # Stores all version 0 aspects for a given createdon timestamp
        # Once we have emitted all aspects for a given timestamp, we can emit the version 0 aspects
        # Guaranteeing that, for a given timestamp, we always ingest version 0 aspects last
        self.queue: Optional[Tuple[datetime, List[ROW]]] = None
        self.enabled = enabled

    def __call__(self, rows: Iterable[ROW]) -> Iterable[ROW]:
        for row in rows:
            yield from self._process_row(row)
        yield from self._flush_queue()

    def _process_row(self, row: ROW) -> Iterable[ROW]:
        if not self.enabled:
            yield row
            return

        yield from self._attempt_queue_flush(row)
        if row["version"] == 0:
            self._add_to_queue(row)
        else:
            yield row

    def _add_to_queue(self, row: ROW) -> None:
        if self.queue is None:
            self.queue = (row["createdon"], [row])
        else:
            self.queue[1].append(row)

    def _attempt_queue_flush(self, row: ROW) -> Iterable[ROW]:
        if self.queue is None:
            return

        if row["createdon"] > self.queue[0]:
            yield from self._flush_queue()

    def _flush_queue(self) -> Iterable[ROW]:
        if self.queue is not None:
            yield from self.queue[1]
            self.queue = None


class DataHubDatabaseReader:
    def __init__(
        self,
        config: DataHubSourceConfig,
        connection_config: SQLAlchemyConnectionConfig,
        report: DataHubSourceReport,
    ):
        self.config = config
        self.report = report
        self.engine = create_engine(
            url=connection_config.get_sql_alchemy_url(),
            **connection_config.options,
        )

        # Cache for available dates to avoid redundant queries
        self.available_dates_cache: Optional[List[datetime]] = None

    @property
    def soft_deleted_urns_query(self) -> str:
        return f"""
            SELECT DISTINCT mav.urn
            FROM {self.engine.dialect.identifier_preparer.quote(self.config.database_table_name)} as mav
            JOIN (
                SELECT *,
                JSON_EXTRACT(metadata, '$.removed') as removed
                FROM {self.engine.dialect.identifier_preparer.quote(self.config.database_table_name)}
                WHERE aspect = "status" AND version = 0
            ) as sd ON sd.urn = mav.urn
            WHERE sd.removed = true
            ORDER BY mav.urn
        """

    def query(self, set_structured_properties_filter: bool) -> str:
        """
        Main query that gets data for specified date range with appropriate filters.
        """
        structured_prop_filter = f" AND urn {'' if set_structured_properties_filter else 'NOT'} like 'urn:li:structuredProperty:%%'"

        return f"""
        SELECT *
        FROM (
            SELECT
                mav.urn,
                mav.aspect,
                mav.metadata,
                mav.systemmetadata,
                mav.createdon,
                mav.version,
                removed
            FROM {self.engine.dialect.identifier_preparer.quote(self.config.database_table_name)} as mav
            LEFT JOIN (
                SELECT
                    *,
                    JSON_EXTRACT(metadata, '$.removed') as removed
                FROM {self.engine.dialect.identifier_preparer.quote(self.config.database_table_name)}
                WHERE aspect = 'status'
                AND version = 0
            ) as sd ON sd.urn = mav.urn
            WHERE 1 = 1
                {"" if self.config.include_all_versions else "AND mav.version = 0"}
                {"" if not self.config.exclude_aspects else "AND mav.aspect NOT IN %(exclude_aspects)s"}
                AND mav.createdon >= %(since_createdon)s
                AND mav.createdon < %(end_createdon)s
            ORDER BY
                createdon,
                urn,
                aspect,
                version
        ) as t
        WHERE 1=1
            {"" if self.config.include_soft_deleted_entities else " AND (removed = false or removed is NULL)"}
            {structured_prop_filter}
        ORDER BY
            createdon,
            urn,
            aspect,
            version
        LIMIT %(limit)s
        OFFSET %(offset)s
        """

    def execute_with_params(
        self, query: str, params: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """Execute query with proper parameter binding that works with your database"""
        with self.engine.connect() as conn:
            result = conn.execute(query, params or {})
            return [dict(row) for row in result.fetchall()]

    def execute_server_cursor(
        self, query: str, params: Dict[str, Any]
    ) -> Iterable[Dict[str, Any]]:
        """Execute a query with server-side cursor"""
        with self.engine.connect() as conn:
            if self.engine.dialect.name in ["postgresql", "mysql", "mariadb"]:
                with (
                    conn.begin()
                ):  # Transaction required for PostgreSQL server-side cursor
                    # Set query timeout at the connection level
                    if self.config.query_timeout:
                        if self.engine.dialect.name == "postgresql":
                            conn.execute(
                                text(
                                    f"SET statement_timeout = {self.config.query_timeout * 1000}"
                                )
                            )  # milliseconds
                        elif self.engine.dialect.name in ["mysql", "mariadb"]:
                            conn.execute(
                                text(
                                    f"SET max_execution_time = {self.config.query_timeout * 1000}"
                                )
                            )  # milliseconds

                    # Stream results with batch size
                    conn = conn.execution_options(
                        stream_results=True,
                        yield_per=self.config.database_query_batch_size,
                    )

                    # Execute query - using native parameterization without text()
                    # to maintain compatibility with your original code
                    result = conn.execute(query, params)
                    for row in result:
                        yield dict(row)

                return  # Success, exit the retry loop
            else:
                raise ValueError(f"Unsupported dialect: {self.engine.dialect.name}")

    def _get_rows(
        self,
        start_date: datetime,
        end_date: datetime,
        set_structured_properties_filter: bool,
        limit: int,
    ) -> Iterable[Dict[str, Any]]:
        """
        Retrieves data rows within a specified date range using pagination.

        Implements a hybrid pagination strategy that switches between time-based and
        offset-based approaches depending on the returned data. Uses server-side
        cursors for efficient memory usage.

        Note: May return duplicate rows across batch boundaries when multiple rows
        share the same 'createdon' timestamp. This is expected behavior when
        transitioning between pagination methods.

        Args:
        start_date: Beginning of date range (inclusive)
        end_date: End of date range (exclusive)
        set_structured_properties_filter: Whether to apply structured filtering
        limit: Maximum rows to fetch per query

        Returns:
            An iterable of database rows as dictionaries
        """
        offset = 0
        last_createdon = None
        first_iteration = True

        while True:
            try:
                # Set up query and parameters - using named parameters
                query = self.query(set_structured_properties_filter)
                params: Dict[str, Any] = {
                    "since_createdon": start_date.strftime(DATETIME_FORMAT),
                    "end_createdon": end_date.strftime(DATETIME_FORMAT),
                    "limit": limit,
                    "offset": offset,
                }

                # Add exclude_aspects if needed
                if (
                    hasattr(self.config, "exclude_aspects")
                    and self.config.exclude_aspects
                ):
                    params["exclude_aspects"] = tuple(self.config.exclude_aspects)

                logger.info(
                    f"Querying data from {start_date.strftime(DATETIME_FORMAT)} to {end_date.strftime(DATETIME_FORMAT)} "
                    f"with limit {limit} and offset {offset} (inclusive range)"
                )

                # Execute query with server-side cursor
                rows = self.execute_server_cursor(query, params)
                # Process and yield rows
                rows_processed = 0
                for row in rows:
                    if first_iteration:
                        start_date = row.get("createdon", start_date)
                        first_iteration = False

                    last_createdon = row.get("createdon")
                    rows_processed += 1
                    yield row

                # If we processed fewer than the limit or no last_createdon, we're done
                if rows_processed < limit or not last_createdon:
                    break

                # Update parameters for next iteration
                if start_date != last_createdon:
                    start_date = last_createdon
                    offset = 0
                else:
                    offset += limit

                logger.info(
                    f"Processed {rows_processed} rows for date range {start_date} to {end_date}. Continuing to next batch."
                )

            except Exception as e:
                logger.error(
                    f"Error processing date range {start_date} to {end_date}: {str(e)}"
                )
                # Re-raise the exception after logging
                raise

    def get_all_aspects(
        self, from_createdon: datetime, stop_time: datetime
    ) -> Iterable[Tuple[MetadataChangeProposalWrapper, datetime]]:
        logger.info("Fetching Structured properties aspects")
        yield from self.get_aspects(
            from_createdon=from_createdon,
            stop_time=stop_time,
            set_structured_properties_filter=True,
        )

        logger.info(
            f"Waiting for {self.config.structured_properties_template_cache_invalidation_interval} seconds for structured properties cache to invalidate"
        )

        time.sleep(
            self.config.structured_properties_template_cache_invalidation_interval
        )

        logger.info("Fetching aspects")
        yield from self.get_aspects(
            from_createdon=from_createdon,
            stop_time=stop_time,
            set_structured_properties_filter=False,
        )

    def get_aspects(
        self,
        from_createdon: datetime,
        stop_time: datetime,
        set_structured_properties_filter: bool = False,
    ) -> Iterable[Tuple[MetadataChangeProposalWrapper, datetime]]:
        orderer = VersionOrderer[Dict[str, Any]](
            enabled=self.config.include_all_versions
        )
        rows = self._get_rows(
            start_date=from_createdon,
            end_date=stop_time,
            set_structured_properties_filter=set_structured_properties_filter,
            limit=self.config.database_query_batch_size,
        )
        for row in orderer(rows):
            mcp = self._parse_row(row)
            if mcp:
                yield mcp, row["createdon"]

    def get_soft_deleted_rows(self) -> Iterable[Dict[str, Any]]:
        """
        Fetches all soft-deleted entities from the database using pagination.

        Yields:
            Row objects containing URNs of soft-deleted entities
        """
        try:
            params: Dict = {}

            logger.debug("Fetching soft-deleted URNs")

            # Use server-side cursor implementation
            rows = self.execute_server_cursor(self.soft_deleted_urns_query, params)
            processed_rows = 0
            # Process and yield rows
            for row in rows:
                processed_rows += 1
                yield row

            logger.debug(f"Fetched batch of {processed_rows} soft-deleted URNs")

        except Exception:
            logger.exception("Error fetching soft-deleted row", exc_info=True)
            raise

    def _parse_row(
        self, row: Dict[str, Any]
    ) -> Optional[MetadataChangeProposalWrapper]:
        try:
            json_aspect = post_json_transform(json.loads(row["metadata"]))
            json_metadata = post_json_transform(
                json.loads(row["systemmetadata"] or "{}")
            )
            system_metadata = SystemMetadataClass.from_obj(json_metadata)
            return MetadataChangeProposalWrapper(
                entityUrn=row["urn"],
                aspect=ASPECT_MAP[row["aspect"]].from_obj(json_aspect),
                systemMetadata=system_metadata,
            )
        except Exception as e:
            logger.warning(
                f"Failed to parse metadata for {row['urn']}: {e}", exc_info=True
            )
            self.report.num_database_parse_errors += 1
            self.report.database_parse_errors.setdefault(
                str(e), LossyDict()
            ).setdefault(row["aspect"], LossyList()).append(row["urn"])
            return None



from enum import Enum


class FieldMetric(Enum):
    UNIQUE_COUNT = "unique_count"
    UNIQUE_PERCENTAGE = "unique_percentage"
    NULL_COUNT = "null_count"
    NULL_PERCENTAGE = "null_percentage"
    MIN = "min"
    MAX = "max"
    MEAN = "mean"
    MEDIAN = "median"
    STDDEV = "stddev"
    NEGATIVE_COUNT = "negative_count"
    NEGATIVE_PERCENTAGE = "negative_percentage"
    ZERO_COUNT = "zero_count"
    ZERO_PERCENTAGE = "zero_percentage"
    MIN_LENGTH = "min_length"
    MAX_LENGTH = "max_length"
    EMPTY_COUNT = "empty_count"
    EMPTY_PERCENTAGE = "empty_percentage"



"""The `RestrictedText` module provides a custom Pydantic type that stores the original
value but returns a truncated and sanitized version when accessed.

Features:
- Configurable maximum length with truncation
- Character replacement (default replaces with underscore)
- Preserves original value internally
- Customizable truncation suffix
- Compatible with both Pydantic v1 and v2
"""

from __future__ import annotations

from typing import Any, ClassVar, Optional, Set, Union

# Check Pydantic version and import accordingly
try:
    from pydantic import VERSION

    PYDANTIC_V2 = int(VERSION.split(".")[0]) >= 2
except (ImportError, AttributeError):
    # Fallback for older versions that don't have VERSION
    PYDANTIC_V2 = False

if PYDANTIC_V2:
    from pydantic import GetCoreSchemaHandler  # type: ignore[attr-defined]
    from pydantic_core import core_schema
else:
    from pydantic.validators import str_validator


class RestrictedTextConfig:
    """Configuration class for RestrictedText."""

    def __init__(
        self,
        max_length: Optional[int] = None,
        forbidden_chars: Optional[Set[str]] = None,
        replacement_char: Optional[str] = None,
        truncation_suffix: Optional[str] = None,
    ):
        self.max_length = max_length
        self.forbidden_chars = forbidden_chars
        self.replacement_char = replacement_char
        self.truncation_suffix = truncation_suffix


class RestrictedText(str):
    """A string type that stores the original value but returns a truncated and sanitized version.

    This type allows you to:
    - Set a maximum length for the displayed value
    - Replace specific characters with a replacement character
    - Access both the original and processed values

    ```python
    from pydantic import BaseModel

    class TestModel(BaseModel):
        # Basic usage with default settings
        name: RestrictedText

        # Custom max length and character replacement using Field
        custom_field: RestrictedText = RestrictedText.with_config(
            max_length=10,
            forbidden_chars={' ', '-', '.'},
            replacement_char='_'
        )

    # Usage example
    model = TestModel(
        name="This is a very long string with special characters!",
        custom_field="hello-world.test"
    )

    print(model.name)  # Truncated and sanitized version
    print(model.name.original)  # Original value
    print(model.custom_field)  # "hello_worl..."
    ```
    """

    # Default configuration
    _default_max_length: ClassVar[Optional[int]] = 50
    _default_forbidden_chars: ClassVar[Set[str]] = {" ", "\t", "\n", "\r"}
    _default_replacement_char: ClassVar[str] = "_"
    _default_truncation_suffix: ClassVar[str] = "..."

    def __new__(cls, value: str = "") -> "RestrictedText":
        """Create a new string instance."""
        instance = str.__new__(cls, "")  # We'll set the display value later
        return instance

    def __init__(self, value: str = ""):
        """Initialize the RestrictedText with a value."""
        self.original: str = value
        self.max_length = self._default_max_length
        self.forbidden_chars = self._default_forbidden_chars
        self.replacement_char = self._default_replacement_char
        self.truncation_suffix = self._default_truncation_suffix

        # Process the value
        self._processed_value = self._process_value(value)

    def _configure(
        self,
        max_length: Optional[int] = None,
        forbidden_chars: Optional[Set[str]] = None,
        replacement_char: Optional[str] = None,
        truncation_suffix: Optional[str] = None,
    ) -> "RestrictedText":
        """Configure this instance with custom settings."""
        if max_length is not None:
            self.max_length = max_length
        if forbidden_chars is not None:
            self.forbidden_chars = forbidden_chars
        if replacement_char is not None:
            self.replacement_char = replacement_char
        if truncation_suffix is not None:
            self.truncation_suffix = truncation_suffix

        # Reprocess the value with new configuration
        self._processed_value = self._process_value(self.original)
        return self

    def _process_value(self, value: str) -> str:
        """Process the value by replacing characters and truncating."""
        # Replace specified characters
        processed = value
        for char in self.forbidden_chars:
            processed = processed.replace(char, self.replacement_char)

        # Truncate if necessary
        if self.max_length is not None and len(processed) > self.max_length:
            if len(self.truncation_suffix) >= self.max_length:
                # If suffix is too long, just truncate without suffix
                processed = processed[: self.max_length]
            else:
                # Truncate and add suffix
                truncate_length = self.max_length - len(self.truncation_suffix)
                processed = processed[:truncate_length] + self.truncation_suffix

        return processed

    def __str__(self) -> str:
        """Return the processed (truncated and sanitized) value."""
        return self._processed_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._processed_value!r})"

    @property
    def processed(self) -> str:
        """Get the processed (truncated and sanitized) value."""
        return self._processed_value

    @classmethod
    def with_config(
        cls,
        max_length: Optional[int] = None,
        forbidden_chars: Optional[Set[str]] = None,
        replacement_char: Optional[str] = None,
        truncation_suffix: Optional[str] = None,
    ) -> RestrictedTextConfig:
        """Create a configuration object for use as field default.

        Args:
            max_length: Maximum length of the processed string
            forbidden_chars: Set of characters to replace
            replacement_char: Character to use as replacement
            truncation_suffix: Suffix to add when truncating

        Returns:
            A configuration object that can be used as field default
        """
        return RestrictedTextConfig(
            max_length=max_length,
            forbidden_chars=forbidden_chars,
            replacement_char=replacement_char,
            truncation_suffix=truncation_suffix,
        )

    # Pydantic v2 methods
    if PYDANTIC_V2:

        @classmethod
        def _validate(
            cls,
            __input_value: Union[str, "RestrictedText"],
            _: core_schema.ValidationInfo,
        ) -> "RestrictedText":
            """Validate and create a RestrictedText instance."""
            if isinstance(__input_value, RestrictedText):
                return __input_value
            return cls(__input_value)

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source: type[Any], handler: GetCoreSchemaHandler
        ) -> core_schema.CoreSchema:
            """Get the Pydantic core schema for this type."""
            return core_schema.with_info_after_validator_function(
                cls._validate,
                core_schema.str_schema(),
                field_name=cls.__name__,
            )

    # Pydantic v1 methods
    else:

        @classmethod
        def __get_validators__(cls):
            """Pydantic v1 validator method."""
            yield cls.validate

        @classmethod
        def validate(cls, v, field=None):
            """Validate and create a RestrictedText instance for Pydantic v1."""
            if isinstance(v, RestrictedText):
                return v

            if not isinstance(v, str):
                # Let pydantic handle the string validation
                v = str_validator(v)

            # Create instance
            instance = cls(v)

            # Check if there's a field default that contains configuration
            if (
                field
                and hasattr(field, "default")
                and isinstance(field.default, RestrictedTextConfig)
            ):
                config = field.default
                instance._configure(
                    max_length=config.max_length,
                    forbidden_chars=config.forbidden_chars,
                    replacement_char=config.replacement_char,
                    truncation_suffix=config.truncation_suffix,
                )

            return instance

        @classmethod
        def __modify_schema__(cls, field_schema):
            """Modify the JSON schema for Pydantic v1."""
            field_schema.update(type="string", examples=["example string"])



import json
import shlex
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

import requests
from requests.auth import HTTPBasicAuth

from datahub.emitter.aspect import JSON_CONTENT_TYPE, JSON_PATCH_CONTENT_TYPE
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.serialization_helper import pre_json_transform
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeProposal,
)
from datahub.metadata.schema_classes import ChangeTypeClass


def _decode_bytes(value: Union[str, bytes]) -> str:
    """Decode bytes to string, if necessary."""
    if isinstance(value, bytes):
        return value.decode()
    return value


def _format_header(name: str, value: Union[str, bytes]) -> str:
    if name == "Authorization":
        return f"{name!s}: <redacted>"
    return f"{name!s}: {_decode_bytes(value)}"


def make_curl_command(
    session: requests.Session, method: str, url: str, payload: Optional[str] = None
) -> str:
    fragments: List[str] = ["curl", "-X", method]

    for header_name, header_value in session.headers.items():
        fragments.extend(["-H", _format_header(header_name, header_value)])

    if session.auth:
        if isinstance(session.auth, HTTPBasicAuth):
            fragments.extend(
                ["-u", f"{_decode_bytes(session.auth.username)}:<redacted>"]
            )
        else:
            # For other auth types, they should be handled via headers
            fragments.extend(["-H", "<unknown auth type>"])

    if payload:
        fragments.extend(["--data", payload])

    fragments.append(url)
    return shlex.join(fragments)


@dataclass
class OpenApiRequest:
    """Represents an OpenAPI request for entity operations."""

    method: str
    url: str
    payload: List[Dict[str, Any]]

    @classmethod
    def from_mcp(
        cls,
        mcp: Union[MetadataChangeProposal, MetadataChangeProposalWrapper],
        gms_server: str,
        async_flag: bool = False,
        search_sync_flag: bool = False,
    ) -> Optional["OpenApiRequest"]:
        """Factory method to create an OpenApiRequest from a MetadataChangeProposal."""
        if not mcp.aspectName or (
            mcp.changeType != ChangeTypeClass.DELETE and not mcp.aspect
        ):
            return None

        method = "post"
        url = f"{gms_server}/openapi/v3/entity/{mcp.entityType}?async={'true' if async_flag else 'false'}"
        payload = []

        if mcp.changeType == ChangeTypeClass.DELETE:
            method = "delete"
            url = f"{gms_server}/openapi/v3/entity/{mcp.entityType}/{mcp.entityUrn}"
        else:
            if mcp.aspect:
                mcp_headers = {}

                if not async_flag and search_sync_flag:
                    mcp_headers["X-DataHub-Sync-Index-Update"] = "true"

                if mcp.changeType == ChangeTypeClass.PATCH:
                    method = "patch"
                    obj = mcp.aspect.to_obj()
                    content_type = obj.get("contentType")
                    if obj.get("value") and content_type == JSON_PATCH_CONTENT_TYPE:
                        # Undo double serialization.
                        obj = json.loads(obj["value"])
                        patch_value = obj
                    else:
                        raise NotImplementedError(
                            f"ChangeType {mcp.changeType} only supports context type {JSON_PATCH_CONTENT_TYPE}, found {content_type}."
                        )

                    if isinstance(patch_value, list):
                        patch_value = {"patch": patch_value}

                    payload = [
                        {
                            "urn": mcp.entityUrn,
                            mcp.aspectName: {
                                "value": patch_value,
                                "systemMetadata": mcp.systemMetadata.to_obj()
                                if mcp.systemMetadata
                                else None,
                                "headers": mcp_headers,
                            },
                        }
                    ]
                else:
                    if isinstance(mcp, MetadataChangeProposalWrapper):
                        aspect_value = pre_json_transform(
                            mcp.to_obj(simplified_structure=True)
                        )["aspect"]["json"]
                    else:
                        obj = mcp.aspect.to_obj()
                        content_type = obj.get("contentType")
                        if obj.get("value") and content_type == JSON_CONTENT_TYPE:
                            # Undo double serialization.
                            obj = json.loads(obj["value"])
                        elif content_type == JSON_PATCH_CONTENT_TYPE:
                            raise NotImplementedError(
                                f"ChangeType {mcp.changeType} does not support patch."
                            )
                        aspect_value = pre_json_transform(obj)

                    payload = [
                        {
                            "urn": mcp.entityUrn,
                            mcp.aspectName: {
                                "value": aspect_value,
                                "systemMetadata": mcp.systemMetadata.to_obj()
                                if mcp.systemMetadata
                                else None,
                                "headers": mcp_headers,
                            },
                        }
                    ]
            else:
                raise ValueError(f"ChangeType {mcp.changeType} requires a value.")

        return cls(method=method, url=url, payload=payload)



from datahub.ingestion.api.registry import PluginRegistry
from datahub.lite.lite_local import DataHubLiteLocal

lite_registry = PluginRegistry[DataHubLiteLocal]()

# We currently only have one implementation.
lite_registry.register_lazy("duckdb", "datahub.lite.duckdb_lite:DuckDBLite")



from dataclasses import dataclass

from tableauserverclient import Server, UserItem

from datahub.ingestion.source.tableau import tableau_constant as c


@dataclass
class UserInfo:
    user_name: str
    site_role: str
    site_id: str

    def has_site_administrator_explorer_privileges(self):
        return self.site_role in [
            c.ROLE_SITE_ADMIN_EXPLORER,
            c.ROLE_SITE_ADMIN_CREATOR,
            c.ROLE_SERVER_ADMIN,
        ]

    @staticmethod
    def from_server(server: Server) -> "UserInfo":
        assert server.user_id, "make the connection with tableau"

        user: UserItem = server.users.get_by_id(server.user_id)

        assert user.site_role, "site_role is not available"  # to silent the lint

        assert user.name, "user name is not available"  # to silent the lint

        assert server.site_id, "site identifier is not available"  # to silent the lint

        return UserInfo(
            user_name=user.name,
            site_role=user.site_role,
            site_id=server.site_id,
        )



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .......schema_classes import ConditionClass
from .......schema_classes import ConjunctiveCriterionClass
from .......schema_classes import CriterionClass
from .......schema_classes import FilterClass


Condition = ConditionClass
ConjunctiveCriterion = ConjunctiveCriterionClass
Criterion = CriterionClass
Filter = FilterClass

# fmt: on



import json
import logging
from typing import List

from datahub.configuration.config_loader import EnvResolver
from datahub.secret.secret_store import SecretStore

logger = logging.getLogger(__name__)


def resolve_secrets(secret_names: List[str], secret_stores: List[SecretStore]) -> dict:
    # Attempt to resolve secret using by checking each configured secret store.
    final_secret_values = dict({})

    for secret_store in secret_stores:
        try:
            # Retrieve secret values from the store.
            secret_values_dict = secret_store.get_secret_values(secret_names)
            # Overlay secret values from each store, if not None.
            for secret_name, secret_value in secret_values_dict.items():
                if secret_value is not None:
                    # HACK: We previously, incorrectly replaced newline characters with
                    # a r'\n' string. This was a lossy conversion, since we can no longer
                    # distinguish between a newline character and the literal '\n' in
                    # the secret value. For now, we assume that all r'\n' strings are
                    # actually newline characters. This will break if a secret value
                    # genuinely contains the string r'\n'.
                    # Once this PR https://github.com/datahub-project/datahub/pull/9484
                    # has baked for a while, we should be able to remove this hack.
                    # TODO: This logic should live in the DataHub secret client/store,
                    # not the general secret resolution logic.
                    secret_value = secret_value.replace(r"\n", "\n")

                    final_secret_values[secret_name] = secret_value
        except Exception:
            logger.exception(
                f"Failed to fetch secret values from secret store with id {secret_store.get_id()}"
            )
    return final_secret_values


def resolve_recipe(
    recipe: str, secret_stores: List[SecretStore], strict_env_syntax: bool = True
) -> dict:
    # Note: the default for `strict_env_syntax` is normally False, but here we override
    # it to be true. Particularly when fetching secrets from external secret stores, we
    # want to be more careful about not over-fetching secrets.

    json_recipe_raw = json.loads(recipe)

    # 1. Extract all secrets needing resolved.
    secrets_to_resolve = EnvResolver.list_referenced_variables(
        json_recipe_raw, strict_env_syntax=strict_env_syntax
    )

    # 2. Resolve secret values
    secret_values_dict = resolve_secrets(list(secrets_to_resolve), secret_stores)

    # 3. Substitute secrets into recipe file
    resolver = EnvResolver(
        environ=secret_values_dict, strict_env_syntax=strict_env_syntax
    )
    json_recipe_resolved = resolver.resolve(json_recipe_raw)

    return json_recipe_resolved



import dataclasses
from collections import defaultdict
from typing import Dict, Iterable, List, Optional

from datahub.configuration.source_common import LowerCaseDatasetUrnConfigMixin
from datahub.configuration.time_window_config import BaseTimeWindowConfig
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import SupportStatus, config_class, support_status
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.snowflake.snowflake_config import (
    SnowflakeFilterConfig,
    SnowflakeIdentifierConfig,
)
from datahub.ingestion.source.snowflake.snowflake_connection import (
    SnowflakeConnectionConfig,
)
from datahub.ingestion.source.snowflake.snowflake_schema import SnowflakeDatabase
from datahub.ingestion.source.snowflake.snowflake_schema_gen import (
    SnowflakeSchemaGenerator,
)
from datahub.ingestion.source.snowflake.snowflake_utils import (
    SnowflakeFilter,
    SnowflakeIdentifierBuilder,
)
from datahub.ingestion.source_report.time_window import BaseTimeWindowReport
from datahub.utilities.lossy_collections import LossyList


class SnowflakeSummaryConfig(
    SnowflakeFilterConfig,
    SnowflakeConnectionConfig,
    BaseTimeWindowConfig,
    LowerCaseDatasetUrnConfigMixin,
):
    pass


@dataclasses.dataclass
class SnowflakeSummaryReport(SourceReport, BaseTimeWindowReport):
    filtered: LossyList[str] = dataclasses.field(default_factory=LossyList)

    num_get_tables_for_schema_queries: int = 0
    num_get_views_for_schema_queries: int = 0

    schema_counters: Dict[str, int] = dataclasses.field(default_factory=dict)
    object_counters: Dict[str, Dict[str, int]] = dataclasses.field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )

    num_snowflake_queries: Optional[int] = None
    num_snowflake_mutations: Optional[int] = None

    def report_dropped(self, ent_name: str) -> None:
        self.filtered.append(ent_name)

    def report_entity_scanned(self, name: str, ent_type: str = "table") -> None:
        pass


@config_class(SnowflakeSummaryConfig)
@support_status(SupportStatus.INCUBATING)
class SnowflakeSummarySource(Source):
    def __init__(self, ctx: PipelineContext, config: SnowflakeSummaryConfig):
        super().__init__(ctx)
        self.config: SnowflakeSummaryConfig = config
        self.report: SnowflakeSummaryReport = SnowflakeSummaryReport()

        self.connection = self.config.get_connection()

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        schema_generator = SnowflakeSchemaGenerator(
            # This is a hack, but we just hope that the config / report have all the fields we need.
            config=self.config,  # type: ignore
            report=self.report,  # type: ignore
            connection=self.connection,
            identifiers=SnowflakeIdentifierBuilder(
                identifier_config=SnowflakeIdentifierConfig(),
                structured_reporter=self.report,
            ),
            domain_registry=None,
            profiler=None,
            aggregator=None,
            snowsight_url_builder=None,
            filters=SnowflakeFilter(
                filter_config=self.config,
                structured_reporter=self.report,
            ),
        )

        # Databases.
        databases: List[SnowflakeDatabase] = []
        for database in schema_generator.get_databases() or []:
            # TODO: Support database_patterns.
            if not self.config.database_pattern.allowed(database.name):
                self.report.report_dropped(f"{database.name}.*")
            else:
                databases.append(database)

        # Schemas.
        for database in databases:
            schema_generator.fetch_schemas_for_database(database, database.name)

            self.report.schema_counters[database.name] = len(database.schemas)

            for schema in database.schemas:
                # Tables/views.
                tables = schema_generator.fetch_tables_for_schema(
                    schema, database.name, schema.name
                )
                views = schema_generator.fetch_views_for_schema(
                    schema, database.name, schema.name
                )

                self.report.object_counters[database.name][schema.name] = len(
                    tables
                ) + len(views)

        # Queries for usage.
        start_time_millis = self.config.start_time.timestamp() * 1000
        end_time_millis = self.config.end_time.timestamp() * 1000
        for row in self.connection.query(
            f"""\
SELECT COUNT(*) AS CNT
FROM snowflake.account_usage.query_history
WHERE query_history.start_time >= to_timestamp_ltz({start_time_millis}, 3)
  AND query_history.start_time < to_timestamp_ltz({end_time_millis}, 3)
"""
        ):
            self.report.num_snowflake_queries = row["CNT"]

        # Queries for lineage/operations.
        for row in self.connection.query(
            f"""\
SELECT COUNT(*) AS CNT
FROM
    snowflake.account_usage.access_history access_history
WHERE query_start_time >= to_timestamp_ltz({start_time_millis}, 3)
  AND query_start_time < to_timestamp_ltz({end_time_millis}, 3)
  AND access_history.objects_modified is not null
  AND ARRAY_SIZE(access_history.objects_modified) > 0
"""
        ):
            self.report.num_snowflake_mutations = row["CNT"]

        # This source doesn't produce any metadata itself. All important information goes into the report.
        yield from []

    def get_report(self) -> SnowflakeSummaryReport:
        return self.report
